# -*- coding: utf-8 -*-
# pylint: disable=R1708
from __future__ import unicode_literals

import io
import logging
import os
import re
import sys
import unittest

from cmakelang import configuration
from cmakelang import lex

from cmakelang.lint import __main__
from cmakelang.lint import lint_util
from cmakelang.lint.test import genfiles


def overzip(iterable_a, iterable_b):
  """
  Like itertools.izip but instead if the two lists have different sizes then
  the resulting generator will yield a number of pairs equal to the larger of
  the two inputs (rathe than the smaller). The empty list will be padded with
  None elements.
  """
  iter_a = iter(iterable_a)
  iter_b = iter(iterable_b)

  item_a = next(iter_a, None)
  item_b = next(iter_b, None)

  # NOTE(josh): this only matters when overzipping a parse tree. It's not
  # meaningful for overzipping a layout tree, but it doesn't hurt since
  # lexer tokens don't show up in the layout tree
  while isinstance(item_a, lex.Token):
    item_a = next(iter_a, None)

  while item_a is not None and item_b is not None:
    yield(item_a, item_b)
    item_a = next(iter_a, None)
    while isinstance(item_a, lex.Token):
      item_a = next(iter_a, None)
    item_b = next(iter_b, None)

  while item_a is not None:
    yield(item_a, None)
    item_a = next(iter_a, None)
    while isinstance(item_a, lex.Token):
      item_a = next(iter_a, None)

  while item_b is not None:
    yield(None, item_b)
    item_b = next(iter_b, None)


def camel_to_snake_callback(match):
  """
  Substituion callback for camel_to_snake
  """
  chars = match.group(0)
  return "{}_{}".format(chars[0], chars[1])


def camel_to_snake(camelstr):
  """
  Translate a camelCaseString into a snake_case_string
  """
  return re.sub("[a-z][A-Z]", camel_to_snake_callback, camelstr).lower()


def snake_to_camel_callback(match):
  """
  Substituion callback for camel_to_snake
  """
  snake_chars = match.group(0)
  return snake_chars[0] + snake_chars[-1].upper()


def snake_to_camel(camelstr, upper=False):
  """
  Translate a camelCaseString into a snake_case_string
  """
  lower_camel = re.sub("([^_]?_+[^_]?)", snake_to_camel_callback, camelstr)
  if upper:
    return lower_camel[0].upper() + lower_camel[1:]
  return lower_camel


class TestBase(unittest.TestCase):
  def __init__(self, *args):
    super(TestBase, self).__init__(*args)
    self.config = configuration.Configuration()

  def execute_test(self, test_name, test_body, expect_list):
    outfile = io.StringIO()
    global_ctx = lint_util.GlobalContext(outfile)
    local_ctx = global_ctx.get_file_ctx(test_name, self.config)
    __main__.process_file(self.config, local_ctx, test_body)
    for actual, expected in overzip(local_ctx.get_lint(), expect_list):
      if expected is None:
        raise AssertionError(
            "More lint than expected, starting with {}".format(actual))

      if "@" in expected:
        expect_id, expect_locstr = expected.split("@")
        expect_loc = tuple(int(val) for val in expect_locstr.split(":"))
      else:
        expect_id = expected
        expect_loc = ()

      if actual is None:
        if expect_loc:
          raise AssertionError(
              "Missing expected lint {} at {}".format(expect_id, expect_loc))

        raise AssertionError(
            "Missing expected lint {}".format(expect_id))

      if expect_id != actual.spec.idstr:
        raise AssertionError(
            "Expected lint {} but got lint {}".format(expect_id, actual))

      actual_loc = actual.location
      if actual_loc is None:
        actual_loc = ()
      for expect_val, actual_val in zip(expect_loc, actual_loc):
        if expect_val != actual_val:
          raise AssertionError(
              "Expected lint {}@{} but got it at {}".format(
                  expect_id, ":".join(str(x) for x in expect_loc),
                  actual_loc))


EXCLUSIONS = [
    "expect_lint.cmake"
]


def iter_testfiles():
  thisdir = os.path.dirname(os.path.realpath(__file__))
  for dirpath, _dirnames, filenames in os.walk(thisdir):
    for filename in filenames:
      if filename in EXCLUSIONS:
        continue
      if filename.endswith(".cmake"):
        yield os.path.join(dirpath, filename)


def iter_tests_from_file(filepath):
  with io.open(filepath, "r", newline='') as infile:
    # Split lines but if the final newline is missing then preserve that fact
    items = re.split("(\n)", infile.read())
    lines = []
    for item in items:
      if item == "\n":
        lines[-1] += item
      else:
        lines.append(item)

  # Simple state-machine parser
  test_name = None
  line_buffer = []
  expect_str = None
  lineiter = enumerate(lines)
  for lineno, line in lineiter:
    if line.startswith("# test: "):
      if line_buffer:
        if test_name is None:
          raise ValueError(
              "Malformed sidecar {}:{}".format(filepath, lineno))
        yield (test_name, "".join(line_buffer), expect_str)
      test_name = line[len("# test: "):].rstrip()
      line_buffer = []
      expect_str = None
    elif line.startswith("# expect:"):
      expect_str = line[len("# expect:"):].strip()
    elif line.endswith("# end-test"):
      if test_name is None:
        raise ValueError(
            "Malformed sidecar {}:{}".format(filepath, lineno))
      yield (test_name, "".join(line_buffer), expect_str)
      test_name = None
      line_buffer = []
      expect_str = None
    else:
      line_buffer.append(line)

  if test_name and line_buffer:
    yield (test_name, "".join(line_buffer), expect_str)


def make_test_fun(test_name, test_body, expect_list):
  def test_fun(self):
    self.execute_test(test_name, test_body, expect_list)
  if sys.version_info < (3, 0, 0):
    # In python 2.7 test_name is a unicode object. We need to convert it to
    # a string.
    test_name = test_name.encode("utf-8")
  test_fun.__name__ = test_name
  test_fun.__doc__ = " ".join(test_name.split("_")[1:])
  return test_fun


def gen_test_classes():
  genfiles.rewrite_lint_tests()
  for filepath in iter_testfiles():
    basename = os.path.splitext(os.path.basename(filepath))[0]
    classname = snake_to_camel(basename, upper=True)
    defn = {}
    for test_name, test_body, expect_str in iter_tests_from_file(filepath):
      method_name = "test_" + test_name.replace("-", "_")
      expect_list = expect_str.rstrip().split(",")
      while expect_list and not expect_list[0]:
        expect_list.pop(0)
      defn[method_name] = make_test_fun(method_name, test_body, expect_list)

    yield type(classname, (TestBase,), defn)


class ConfigTestCase(TestBase):
  """
  Test that config options function correctly
  """

  def test_disabled_codes(self):
    self.config.lint.disabled_codes = ["E1120"]
    test_body = "set(VARNAME varvalue CACHE STRING)\n"
    self.execute_test("test_disabled_code", test_body, [])

  def test_tabpolicy_tabs(self):
    self.config.format.use_tabchars = True
    self.config.format.fractional_tab_policy = "round-up"
    test_body = "if(TRUE)\n\t message(\"hello\")\nendif()\n"
    self.execute_test("test_tabpolicy_tabs", test_body, ["C0306", "C0307"])


classobj = None
for classobj in gen_test_classes():
  globals()[classobj.__name__] = classobj
del classobj

if __name__ == "__main__":
  logging.basicConfig(level=logging.INFO)
  unittest.main()



# -*- coding: utf-8 -*-
# pylint: disable=R1708
from __future__ import unicode_literals
import unittest

from cmakelang import configuration
from cmakelang import lex
from cmakelang import parse
from cmakelang.parse.printer import tree_string, test_string
from cmakelang.parse.common import NodeType


def overzip(iterable_a, iterable_b):
  """
  Like itertools.izip but instead if the two lists have different sizes then
  the resulting generator will yield a number of pairs equal to the larger of
  the two inputs (rathe than the smaller). The empty list will be padded with
  None elements.
  """
  iter_a = iter(iterable_a)
  iter_b = iter(iterable_b)

  item_a = next(iter_a, None)
  item_b = next(iter_b, None)

  while isinstance(item_a, lex.Token):
    item_a = next(iter_a, None)

  while item_a is not None and item_b is not None:
    yield(item_a, item_b)
    item_a = next(iter_a, None)
    while isinstance(item_a, lex.Token):
      item_a = next(iter_a, None)
    item_b = next(iter_b, None)

  while item_a is not None:
    yield(item_a, None)
    item_a = next(iter_a, None)
    while isinstance(item_a, lex.Token):
      item_a = next(iter_a, None)

  while item_b is not None:
    yield(None, item_b)
    item_b = next(iter_b, None)


def assert_tree_type(test, nodes, tups, tree=None, history=None):
  """
  Check the output tree structure against that of expect_tree: a nested tuple
  tree.
  """

  if tree is None:
    tree = nodes

  if history is None:
    history = []

  for node, tup in overzip(nodes, tups):
    if isinstance(node, lex.Token):
      continue
    message = ("For node {} at\n {} within \n{}. "
               "If this is infact correct, copy-paste this:\n\n{}"
               .format(node, tree_string([node]),
                       tree_string(tree, history),
                       test_string(tree)))
    test.assertIsNotNone(node, msg="Missing node " + message)
    test.assertIsNotNone(tup, msg="Extra node " + message)
    expect_type, expect_children = tup
    test.assertEqual(node.node_type, expect_type,
                     msg="Expected type={} ".format(expect_type) + message)
    assert_tree_type(test, node.children, expect_children, tree,
                     history + [node])


class TestCanonicalParse(unittest.TestCase):
  """
  Given a bunch of example inputs, ensure that they parse into the expected
  tree structure.
  """

  def __init__(self, *args, **kwargs):
    super(TestCanonicalParse, self).__init__(*args, **kwargs)
    self.config = configuration.Configuration()
    parse_db = parse.funs.get_parse_db()
    self.parse_ctx = parse.ParseContext(parse_db)

  def setUp(self):
    self.config.parse.fn_spec.add(
        'foo',
        flags=['BAR', 'BAZ'],
        kwargs={
            "HEADERS": '*',
            "SOURCES": '*',
            "DEPENDS": '*'
        })

    self.parse_ctx.parse_db.update(
        parse.funs.get_funtree(self.config.parse.fn_spec))

  def do_type_test(self, input_str, expect_tree):
    """
    Run the parser to get the fst, then compare the result to the types in the
    ``expect_tree`` tuple tree.
    """
    tokens = lex.tokenize(input_str)

    fst_root = parse.parse(tokens, self.parse_ctx)
    assert_tree_type(self, [fst_root], expect_tree)

  def test_collapse_additional_newlines(self):
    self.do_type_test("""\
      # The following multiple newlines should be collapsed into a single newline




      cmake_minimum_required(VERSION 2.8.11)
      project(cmakelang_test)
      """, [
          (NodeType.BODY, [
              (NodeType.WHITESPACE, []),
              (NodeType.COMMENT, []),
              (NodeType.WHITESPACE, []),
              (NodeType.STATEMENT, [
                  (NodeType.FUNNAME, []),
                  (NodeType.LPAREN, []),
                  (NodeType.ARGGROUP, [
                      (NodeType.KWARGGROUP, [
                          (NodeType.KEYWORD, []),
                          (NodeType.ARGGROUP, [
                              (NodeType.PARGGROUP, [
                                  (NodeType.ARGUMENT, []),
                              ]),
                          ]),
                      ]),
                  ]),
                  (NodeType.RPAREN, []),
              ]),
              (NodeType.WHITESPACE, []),
              (NodeType.STATEMENT, [
                  (NodeType.FUNNAME, []),
                  (NodeType.LPAREN, []),
                  (NodeType.ARGGROUP, [
                      (NodeType.PARGGROUP, [
                          (NodeType.ARGUMENT, []),
                      ]),
                  ]),
                  (NodeType.RPAREN, []),
              ]),
              (NodeType.WHITESPACE, []),
          ]),
      ])

  def test_multiline_comment(self):
    self.do_type_test("""\
      # This multiline-comment should be reflowed
      # into a single comment
      # on one line
      """, [
          (NodeType.BODY, [
              (NodeType.WHITESPACE, []),
              (NodeType.COMMENT, []),
              (NodeType.WHITESPACE, []),
          ])
      ])

  def test_nested_kwargs(self):
    self.do_type_test("""\
      add_custom_target(name ALL VERBATIM
        COMMAND echo hello world
        COMMENT "this is some text")
      """, [
          (NodeType.BODY, [
              (NodeType.WHITESPACE, []),
              (NodeType.STATEMENT, [
                  (NodeType.FUNNAME, []),
                  (NodeType.LPAREN, []),
                  (NodeType.ARGGROUP, [
                      (NodeType.PARGGROUP, [
                          (NodeType.ARGUMENT, []),
                          (NodeType.FLAG, []),
                      ]),
                      (NodeType.PARGGROUP, [
                          (NodeType.FLAG, []),
                      ]),
                      (NodeType.KWARGGROUP, [
                          (NodeType.KEYWORD, []),
                          (NodeType.ARGGROUP, [
                              (NodeType.PARGGROUP, [
                                  (NodeType.ARGUMENT, []),
                                  (NodeType.ARGUMENT, []),
                                  (NodeType.ARGUMENT, []),
                              ]),
                          ]),
                      ]),
                      (NodeType.KWARGGROUP, [
                          (NodeType.KEYWORD, []),
                          (NodeType.PARGGROUP, [
                              (NodeType.ARGUMENT, []),
                          ]),
                      ]),
                  ]),
                  (NodeType.RPAREN, []),
              ]),
              (NodeType.WHITESPACE, []),
          ]),
      ])

  def test_custom_command(self):
    self.do_type_test("""\
      # This very long command should be broken up along keyword arguments
      foo(nonkwarg_a nonkwarg_b HEADERS a.h b.h c.h d.h e.h f.h SOURCES a.cc b.cc d.cc DEPENDS foo bar baz)
      """, [
          (NodeType.BODY, [
              (NodeType.WHITESPACE, []),
              (NodeType.COMMENT, []),
              (NodeType.WHITESPACE, []),
              (NodeType.STATEMENT, [
                  (NodeType.FUNNAME, []),
                  (NodeType.LPAREN, []),
                  (NodeType.ARGGROUP, [
                      (NodeType.PARGGROUP, [
                          (NodeType.ARGUMENT, []),
                          (NodeType.ARGUMENT, []),
                      ]),
                      (NodeType.KWARGGROUP, [
                          (NodeType.KEYWORD, []),
                          (NodeType.ARGGROUP, [
                              (NodeType.PARGGROUP, [
                                  (NodeType.ARGUMENT, []),
                                  (NodeType.ARGUMENT, []),
                                  (NodeType.ARGUMENT, []),
                                  (NodeType.ARGUMENT, []),
                                  (NodeType.ARGUMENT, []),
                                  (NodeType.ARGUMENT, []),
                              ]),
                          ]),
                      ]),
                      (NodeType.KWARGGROUP, [
                          (NodeType.KEYWORD, []),
                          (NodeType.ARGGROUP, [
                              (NodeType.PARGGROUP, [
                                  (NodeType.ARGUMENT, []),
                                  (NodeType.ARGUMENT, []),
                                  (NodeType.ARGUMENT, []),
                              ]),
                          ]),
                      ]),
                      (NodeType.KWARGGROUP, [
                          (NodeType.KEYWORD, []),
                          (NodeType.ARGGROUP, [
                              (NodeType.PARGGROUP, [
                                  (NodeType.ARGUMENT, []),
                              ]),
                          ]),
                      ]),
                      (NodeType.PARGGROUP, [
                          (NodeType.FLAG, []),
                          (NodeType.FLAG, []),
                      ]),
                  ]),
                  (NodeType.RPAREN, []),
              ]),
              (NodeType.WHITESPACE, []),
          ]),
      ])

  def test_shellcommand_parse(self):
    self.do_type_test("""\
      add_test(NAME foo-test
               COMMAND cmdname -Bm -h --hello foo bar --world baz buck
               WORKING_DIRECTORY ${CMAKE_SOURCE_DIR})
      """, [
          (NodeType.BODY, [
              (NodeType.WHITESPACE, []),
              (NodeType.STATEMENT, [
                  (NodeType.FUNNAME, []),
                  (NodeType.LPAREN, []),
                  (NodeType.ARGGROUP, [
                      (NodeType.KWARGGROUP, [
                          (NodeType.KEYWORD, []),
                          (NodeType.ARGGROUP, [
                              (NodeType.PARGGROUP, [
                                  (NodeType.ARGUMENT, []),
                              ]),
                          ]),
                      ]),
                      (NodeType.KWARGGROUP, [
                          (NodeType.KEYWORD, []),
                          (NodeType.ARGGROUP, [
                              (NodeType.PARGGROUP, [
                                  (NodeType.ARGUMENT, []),
                                  (NodeType.ARGUMENT, []),
                                  (NodeType.ARGUMENT, []),
                                  (NodeType.ARGUMENT, []),
                                  (NodeType.ARGUMENT, []),
                                  (NodeType.ARGUMENT, []),
                                  (NodeType.ARGUMENT, []),
                                  (NodeType.ARGUMENT, []),
                                  (NodeType.ARGUMENT, []),
                              ]),
                          ]),
                      ]),
                      (NodeType.KWARGGROUP, [
                          (NodeType.KEYWORD, []),
                          (NodeType.ARGGROUP, [
                              (NodeType.PARGGROUP, [
                                  (NodeType.ARGUMENT, []),
                              ]),
                          ]),
                      ]),
                  ]),
                  (NodeType.RPAREN, []),
              ]),
              (NodeType.WHITESPACE, []),
          ]),
      ])


if __name__ == '__main__':
  unittest.main()




import logging

from cmakelang import lex
from cmakelang.parse.common import KwargBreaker, NodeType
from cmakelang.parse.common import TreeNode
from cmakelang.parse.argument_nodes import (
    ArgGroupNode,
    PositionalParser,
    StandardArgTree,
    PositionalGroupNode,
    KeywordGroupNode)
from cmakelang.parse.util import (
    WHITESPACE_TOKENS,
    get_normalized_kwarg,
    get_tag,
    iter_semantic_tokens,
    should_break,
)

logger = logging.getLogger(__name__)


def parse_foreach_standard(ctx, tokens, breakstack):
  """
  ::

    foreach(<loop_var> <items>)
      <commands>
    endforeach()
  """
  tree = ArgGroupNode()

  # If it is a whitespace token then put it directly in the parse tree at
  # the current depth
  while tokens and tokens[0].type in WHITESPACE_TOKENS:
    tree.children.append(tokens.pop(0))
    continue

  # Consume the loop variable and any attached comments
  ntokens = len(tokens)
  subtree = PositionalGroupNode.parse(ctx, tokens, 1, [], breakstack)
  assert len(tokens) < ntokens
  tree.children.append(subtree)

  # Consume the items of iteration
  ntokens = len(tokens)
  subtree = PositionalGroupNode.parse(ctx, tokens, '+', [], breakstack)
  assert len(tokens) < ntokens
  tree.children.append(subtree)

  return tree


def parse_foreach_range(ctx, tokens, breakstack):
  """
  ::

    foreach(<loop_var> RANGE <start> <stop> [<step>])
      <commands>
    endforeach()
  """

  return StandardArgTree.parse(
      ctx, tokens,
      npargs=1,
      kwargs={"RANGE": PositionalParser(3)},
      flags=[],
      breakstack=breakstack
  )


def parse_foreach_in(ctx, tokens, breakstack):
  """
  ::

    foreach(loop_var IN [LISTS [<lists>]] [ITEMS [<items>]])
  """
  tree = ArgGroupNode()

  # If it is a whitespace token then put it directly in the parse tree at
  # the current depth
  while tokens and tokens[0].type in WHITESPACE_TOKENS:
    tree.children.append(tokens.pop(0))
    continue

  # Consume the loop variable and any attached comments
  ntokens = len(tokens)
  pargs = PositionalGroupNode.parse(ctx, tokens, 2, ["IN"], breakstack)
  assert len(tokens) < ntokens
  tree.children.append(pargs)

  kwargs = {
      "LISTS": PositionalParser("+"),
      "ITEMS": PositionalParser("+")
  }
  sub_breakstack = breakstack + [KwargBreaker(list(kwargs.keys()))]

  while tokens:
    # Break if the next token belongs to a parent parser, i.e. if it
    # matches a keyword argument of something higher in the stack, or if
    # it closes a parent group.
    if should_break(tokens[0], breakstack):
      break

    # If it is a whitespace token then put it directly in the parse tree at
    # the current depth
    if tokens[0].type in WHITESPACE_TOKENS:
      tree.children.append(tokens.pop(0))
      continue

    # If it's a comment, then add it at the current depth
    if tokens[0].type in (lex.TokenType.COMMENT,
                          lex.TokenType.BRACKET_COMMENT):
      if not get_tag(tokens[0]) in ("sort", "sortable"):
        child = TreeNode(NodeType.COMMENT)
        tree.children.append(child)
        child.children.append(tokens.pop(0))
        continue

    ntokens = len(tokens)
    word = get_normalized_kwarg(tokens[0])
    subparser = kwargs.get(word, None)
    if subparser is not None:
      subtree = KeywordGroupNode.parse(
          ctx, tokens, word, subparser, sub_breakstack)
    else:
      logger.warning(
          "Unexpected positional argument at %s", tokens[0].location)
      subtree = PositionalGroupNode.parse(ctx, tokens, '+', [], sub_breakstack)

    assert len(tokens) < ntokens
    tree.children.append(subtree)
  return tree


def parse_foreach(ctx, tokens, breakstack):
  """
  ``parse_foreach()`` has a couple of forms:

  * the usual form
  * range form
  * in (lists/items) form

  This function is just the dispatcher

  :see: https://cmake.org/cmake/help/latest/command/foreach.html
  """

  semantic_iter = iter_semantic_tokens(tokens)
  # first token is always the loop variable
  _ = next(semantic_iter, None)
  # Second token is usually the descriminator
  second_token = next(semantic_iter, None)

  if second_token is None:
    # All foreach() statements should have at least two arguments
    logger.warning("Invalid foreach() statement at %s",
                   tokens[0].get_location())
    return StandardArgTree.parse(
        ctx, tokens, npargs='*', kwargs={}, flags=[], breakstack=breakstack)

  descriminator = second_token.spelling.upper()
  dispatchee = {
      "RANGE": parse_foreach_range,
      "IN": parse_foreach_in
  }.get(descriminator, None)
  if dispatchee is not None:
    return dispatchee(ctx, tokens, breakstack)

  return parse_foreach_standard(ctx, tokens, breakstack)


def populate_db(parse_db):
  parse_db["foreach"] = parse_foreach



"""
Check ~/.pbuilderrc for some common advisable configurations
"""

import io
import os
import logging
import re

logger = logging.getLogger(__name__)


def get_extra_packages(content):
  extra_packages = []
  for line in content.split("\n"):
    line = line.strip()
    prefix = "EXTRAPACKAGES="
    if line.startswith(prefix):
      extra_packages.extend(line[len(prefix):].split(","))
  return extra_packages


def get_ldpreloads(content):
  preloads = []
  for line in content.split("\n"):
    line = line.strip()
    prefix = "export LD_PRELOAD="
    if line.startswith(prefix):
      value = line[len(prefix):].strip('"')
      value = re.sub(r"\$\{[^\}]*\}", "", value)
      preloads.extend(value.split(":"))
  return preloads


def main():
  filepath = os.path.expanduser("~/.pbuilderrc")
  if not os.path.exists(filepath):
    logger.warning(
        "No ~/.pbuilderrc in current build environment. It is highly"
        " recommended to include a ~/.pbuilderrc.")
    return

  with io.open(filepath, "r", encoding="utf-8") as infile:
    content = infile.read()

  extra_packages = get_extra_packages(content)
  if "eatmydata" not in extra_packages:
    logger.warning(".pbuilderrc missing EXTRAPACKAGES=eatmydata")

  ld_preloads = get_ldpreloads(content)
  if "libeatmydata.so" not in ld_preloads:
    logger.warning(
        '.pbuilderrc missing '
        '`export LD_PRELOAD="${LD_PRELOAD:+$LD_PRELOAD:}libeatmydata.so"`\n'
        'Contains: %s', ", ".join(ld_preloads))


if __name__ == "__main__":
  main()



"""
Modify a cmake-generated build system to insert our own every-run hook
into the generated build.
"""

import argparse
import io
import os
import resource
import sys
import time


def rewrite_ninja_build(binarydir):
  infilepath = os.path.join(binarydir, "build.ninja")
  outfilepath = infilepath + ".tmp"

  with io.open(outfilepath, "w", encoding="utf-8") as outfile:
    with io.open(infilepath, "r", encoding="utf-8") as infile:
      lineiter = iter(infile)
      for line in lineiter:
        outfile.write(line)
        if line.strip() == "# A missing CMake input file is not an error.":
          break

      # The next line should be a blank line
      outfile.write(next(lineiter, None))

      # Then this is the line we need to modify
      line = next(lineiter, None)
      line = line.replace("codestyle_manifest.cmake ", "")
      outfile.write(line)

      for line in lineiter:
        outfile.write(line)
  os.rename(outfilepath, infilepath)


MAKEFILE_RECIPE = (
    "\tcd $(CMAKE_SOURCE_DIR) && python -B"
    " {TANGENT_TOOLING}/generate_style_manifest.py"
    " --outfile $(CMAKE_BINARY_DIR)/codestyle_manifest.cmake"
    " --excludes-from $(CMAKE_SOURCE_DIR)/build_tooling/lint-excludes.txt"
    " -- $(CMAKE_SOURCE_DIR)\n")


def rewrite_makefiles(binarydir):
  infilepath = os.path.join(binarydir, "CMakeFiles/Makefile2")
  outfilepath = infilepath + ".tmp"

  with io.open(outfilepath, "w", encoding="utf-8") as outfile:
    with io.open(infilepath, "r", encoding="utf-8") as infile:
      lineiter = iter(infile)
      for line in lineiter:
        outfile.write(line)
        if line.strip() == "cmake_check_build_system:":
          break

      outfile.write(MAKEFILE_RECIPE.format(
          TANGENT_TOOLING=os.path.dirname(__file__)))

      for line in lineiter:
        outfile.write(line)
  os.rename(outfilepath, infilepath)


def daemon_main(args):
  if args.generator == "Ninja":
    rewrite_ninja_build(args.binary_dir)
  elif args.generator == "Unix Makefiles":
    rewrite_makefiles(args.binary_dir)


def main():
  argparser = argparse.ArgumentParser(description=__doc__)
  argparser.add_argument("--test-mode", action="store_true")
  argparser.add_argument(
      "--binary-dir", required=True)
  argparser.add_argument(
      "--generator", required=True, choices=["Ninja", "Unix Makefiles"])
  args = argparser.parse_args()

  cmake_pid = os.getppid()
  process_id = os.fork()
  if process_id != 0:
    # this is the parent:
    sys.exit(0)

  # group_id = os.setpgrp()
  process_id = os.setsid()
  if process_id == -1:
    sys.exit(1)

  for fd in range(3, resource.getrlimit(resource.RLIMIT_NOFILE)[0]):
    try:
      os.close(fd)
    except OSError:
      pass

  devnull_fd = os.open("/dev/null", os.O_RDWR)
  os.dup2(devnull_fd, 0)
  os.dup2(devnull_fd, 1)
  os.dup2(devnull_fd, 2)
  os.close(devnull_fd)
  # wait for cmake to die
  while os.path.exists("/proc/{}".format(cmake_pid)):
    if args.test_mode:
      break
    time.sleep(0.5)

  daemon_main(args)


if __name__ == "__main__":
  main()



"""
Scan the source tree rooted at a particulary directory and build a list of all
lintable files. Compare that list against the list of names in any
`.lint-manifest` in the binary tree. If the two are not identical, exit with
nonzero status code.
"""

from __future__ import absolute_import, unicode_literals

import argparse
import difflib
import io
import os
import re
import shlex
import sys

SLUGS = ["BZL", "CMAKE", "CC", "JS", "PY", "SHELL"]


def invert_map(sourcemap):
  out = {}
  for slug, extensions in sourcemap.items():
    for extension in extensions:
      out[extension] = slug
  return out


KNOWN_EXTENSIONS = {
    "BZL": [".bzl"],
    "CMAKE": [".cmake"],
    "CC": [".c", ".cc", ".cxx", ".cpp", ".h", ".hxx", ".hpp", ".tcc"],
    "JS": [".js"],
    "PY": [".py"],
    "SHELL": [".sh"]
}

EXTENSION_MAP = invert_map(KNOWN_EXTENSIONS)


def get_slug_from_extension(filename):
  tail = filename
  if os.sep in tail:
    tail = tail.rsplit(os.sep, 1)[1]

  parts = tail.split(".")
  while parts:
    parts[0] = ""
    candidate_extension = ".".join(parts).lower()
    slug = EXTENSION_MAP.get(candidate_extension, None)
    if slug is not None:
      return slug
    parts.pop(0)

  return None


def get_slug_from_shebang(filepath):
  with io.open(filepath, "rb") as infile:
    shebang = infile.read(2)
    if shebang != b"#!":
      return None
    interp = infile.readline().strip().decode("utf-8")
  parts = shlex.split(interp)
  if len(parts) == 2:
    if parts[0] == "/usr/bin/env":
      if parts[1] == "python":
        return "PY"
      elif parts[1] == "bash":
        return "SHELL"
  else:
    if interp.startswith("/usr/bin/python"):
      return "PY"
    if interp in ["/bin/bash", "/bin/sh"]:
      return "SHELL"
  return None


KNOWN_NAMES = {
    "BZL": ["BUILD", "WORKSPACE"],
    "CMAKE": ["CMakeLists.txt"]
}

NAME_MAP = invert_map(KNOWN_NAMES)


def get_slug_from_name(filename):
  return NAME_MAP.get(filename, None)


REJECT_NAMES = [
    # Outputs of cmake in-source configure
    "CTestTestfile.cmake",
    "cmake_install.cmake",
]


def is_known_reject(filename):
  return filename in REJECT_NAMES


REJECT_DIRS = [
    "build",
    ".git",
    ".build",
    "__pycache__",
]


def get_source_manifest(rootdir, exclude_pattern):
  manifest = {slug: [] for slug in SLUGS}
  excluded = []

  for directory, dirnames, filenames in os.walk(rootdir):
    dirnames[:] = sorted(dirnames)
    relpath_dir = os.path.relpath(directory, rootdir)

    if os.sep in directory:
      dirname = directory.rsplit(os.sep, 1)[1]
    else:
      dirname = directory

    # NOTE(josh): on buildbot, the source directory is called "build"... ugh.
    if dirname in REJECT_DIRS and directory != rootdir:
      excluded.append(relpath_dir)
      dirnames[:] = []
      continue

    if exclude_pattern and exclude_pattern.match(relpath_dir):
      excluded.append(relpath_dir)
      dirnames[:] = []
      continue

    for filename in filenames:
      if is_known_reject(filename):
        continue

      relpath_file = os.path.join(relpath_dir, filename)
      fullpath_file = os.path.join(directory, filename)

      if os.path.islink(fullpath_file):
        continue

      if relpath_dir == ".":
        relpath_file = filename

      if exclude_pattern and exclude_pattern.match(relpath_file):
        excluded.append(relpath_file)
        continue

      slug = get_slug_from_extension(filename)
      if slug is None:
        slug = get_slug_from_name(filename)
      if slug is None:
        slug = get_slug_from_shebang(fullpath_file)

      if slug is not None:
        manifest[slug].append(relpath_file)

  return manifest, excluded


def write_manifest(outfile, manifest, excluded):
  outfile.write("format_and_lint(main\n")
  for slug, filenames in sorted(manifest.items()):
    if not filenames:
      continue
    outfile.write("  {}\n".format(slug))
    for filename in sorted(filenames):
      outfile.write("    {}\n".format(filename))
  outfile.write(")\n")

  if not excluded:
    return

  outfile.write("\n")
  outfile.write("NOTE: The following were excluded\n")
  for filename in sorted(excluded):
    outfile.write("# {}\n".format(filename))


DEFAULT_EXCLUDES = [
    r".*\.egginfo",
]


def main():
  parser = argparse.ArgumentParser(description=__doc__)
  parser.add_argument(
      "-o", "--outfile-path", default="-",
      help="file to write the generated manifest")
  parser.add_argument(
      "--show-exclusions", action="store_true",
      help="include a list of excluded files in the output")

  parser.add_argument("--excludes", nargs="*", help="exclusions")
  parser.add_argument("--excludes-from", help="exclusions file")
  parser.add_argument("source_tree", help="path to the source directory")
  args = parser.parse_args()

  excludes = list(DEFAULT_EXCLUDES)
  if args.excludes:
    excludes.extend(args.exclude)
  if args.excludes_from:
    with io.open(args.excludes_from, "r", encoding="utf-8") as infile:
      for line in infile:
        line = line.strip()
        if not line:
          continue
        if line.startswith("#"):
          continue
        excludes.append(line)

  exclude_pattern = "|".join(excludes)
  exclude_pattern = re.compile(exclude_pattern)

  manifest, excluded = get_source_manifest(args.source_tree, exclude_pattern)

  buffer = io.StringIO()
  if not args.show_exclusions:
    excluded = []
  write_manifest(buffer, manifest, excluded)
  outfile_content = buffer.getvalue()

  if args.outfile_path == "-":
    sys.stdout.write(outfile_content)
    return 0

  if os.path.isfile(args.outfile_path):
    with io.open(args.outfile_path, "r", encoding="utf-8") as infile:
      existing_content = infile.read()
    if existing_content == outfile_content:
      return 0

    for line in difflib.unified_diff(
        existing_content.split("\n"), outfile_content.split("\n")):
      line = line.rstrip("\n")
      print(line)

  with io.open(args.outfile_path, "w", encoding="utf-8") as outfile:
    outfile.write(outfile_content)


if __name__ == "__main__":
  sys.exit(main())



# -*- coding: utf-8 -*-
"""
Parse ctest testfiles and re-emit the test specification in a more structured
format.
"""


import argparse
import collections
import io
import json
import logging
import os
import sys
from xml.etree import ElementTree as ET

from cmakelang.format import __main__
from cmakelang import lex
from cmakelang import parse


logger = logging.getLogger(__name__)


class TestSpec(object):
  """Stores agregated ctest specification, including the name, command, and
     any properties that are set.
  """

  def __init__(self, name, argv, cwd):
    self.name = name
    self.argv = argv
    self.cwd = cwd
    self.props = {}

  def as_odict(self):
    """Return a dictionary representation of the specification, suitable for
       serialization to JSON.
    """
    props = collections.OrderedDict()
    for key, value in sorted(self.props.items()):
      if key == "LABELS":
        value = value.split(";")
      if key == "TIMEOUT":
        value = int(value)
      if key.startswith("_"):
        continue
      props[key.lower()] = value

    out = collections.OrderedDict([
        ("name", self.name),
        ("argv", self.argv),
        ("cwd", self.cwd),
        ("props", props),
    ])
    return out

  def as_element(self):
    """Return an ElementTree representation of the specification, suitable
       for serialialization to XML.
    """
    elem = ET.Element("test")
    elem.set("name", self.name)
    elem.set("cwd", self.cwd)

    argv_elem = ET.Element("argv")
    elem.append(argv_elem)

    for value in self.argv:
      item = ET.Element("arg")
      argv_elem.append(item)
      item.set("value", value)

    for key, value in sorted(self.props.items()):
      if key.startswith("_"):
        continue

      if key == "LABELS":
        labels_elem = ET.Element("labels")
        elem.append(labels_elem)
        for label in value.split(";"):
          item = ET.Element("label")
          labels_elem.append(item)
          item.set("value", label)
        continue

      elem.set(key.lower(), value)
    return elem


class ParseContext(object):
  """Persistent context for testfile parsing. Stores a queue of directories
     to process, and callbacks for each of the different cmake statements.
  """

  def __init__(self):
    self.cwd = ""
    self.dirqueue = []
    self.tests = {}

  # pylint: disable=W0613
  def parse_add_test(self, tokens, _breakstack):
    """Parse an add_test() statement. This statement contains the test name
       and the command for the test.
    """
    if not tokens:
      raise RuntimeError("Ran out of tokens while processing add_test")
    if tokens[0].type not in (
        lex.TokenType.WORD, lex.TokenType.UNQUOTED_LITERAL):
      raise RuntimeError("Expected a WORD token but got {}".format(tokens[0]))
    test_name = tokens[0].spelling
    tokens.pop(0)
    test_argv = []
    while tokens and tokens[0].type != lex.TokenType.RIGHT_PAREN:
      token = tokens.pop(0)
      if token.type is lex.TokenType.WHITESPACE:
        continue
      if token.type is lex.TokenType.QUOTED_LITERAL:
        spelling = token.spelling[1:-1]
      else:
        spelling = token.spelling
      test_argv.append(spelling)

    logger.debug("Adding test %s", test_name)
    self.tests[test_name] = TestSpec(test_name, test_argv, self.cwd)

  def parse_set_tests_properties(self, tokens, _breakstack):
    """Parse a set_tests_properties() statement. This statement can set
       properties (key/value strings) on one or more tests.
    """
    test_names = []
    properties = {}

    while (tokens
           and tokens[0].spelling != "PROPERTIES"
           and tokens[0].type != lex.TokenType.RIGHT_PAREN):
      token = tokens.pop(0)
      if token.type == lex.TokenType.WHITESPACE:
        continue
      assert token.type in (
          lex.TokenType.WORD, lex.TokenType.UNQUOTED_LITERAL), (
              "Unexpected {}".format(token))
      test_names.append(token.spelling)

    token = tokens.pop(0)
    assert token.spelling == "PROPERTIES"

    while tokens and tokens[0].type != lex.TokenType.RIGHT_PAREN:
      token = tokens.pop(0)
      if token.type is lex.TokenType.WHITESPACE:
        token = tokens.pop(0)
      assert token.type in (
          lex.TokenType.WORD, lex.TokenType.UNQUOTED_LITERAL), (
              "Unexpected {}".format(token))
      key = token.spelling

      token = tokens.pop(0)
      if token.type is lex.TokenType.WHITESPACE:
        token = tokens.pop(0)

      if token.type is lex.TokenType.QUOTED_LITERAL:
        value = token.spelling[1:-1]
      else:
        value = token.spelling

      properties[key] = value

    for test_name in test_names:
      logger.debug("Updating properties for %s", test_name)
      self.tests[test_name].props.update(properties)

  def parse_subdirs(self, tokens, _breakstack):
    """Parse a subdirs() statement. This statement is actually deprecated in
       cmake but it appears it is still in use by ctest.
    """
    while tokens and tokens[0].type != lex.TokenType.RIGHT_PAREN:
      token = tokens.pop(0)
      if token.type is lex.TokenType.QUOTED_LITERAL:
        spelling = token.spelling[1:-1]
      else:
        spelling = token.spelling
      fullpath = os.path.join(self.cwd, spelling)
      logger.debug("pushing directory %s", fullpath)
      self.dirqueue.append(fullpath)

  def get_db(self):
    """Return the parse database: dictionary of statement names to parse
       functions. While cmake-format uses regular functions we map statement
       names to callbacks within this context.
    """
    return {
        "add_test": self.parse_add_test,
        "subdirs": self.parse_subdirs,
        "set_tests_properties": self.parse_set_tests_properties,
    }

  def parse_file(self, filepath):
    """Parse one file. Read the content, tokenize, and parse.
    """
    if not os.path.exists(filepath):
      logger.warning("%s does not exist", filepath)
      return
    with io.open(filepath, "r", encoding="utf-8") as infile:
      infile_content = infile.read()
    tokens = lex.tokenize(infile_content)
    _ = parse.parse(tokens, self.get_db())

  def start(self, firstdir):
    """Main entry-point into the parse. Pushes the first directory onto the
       queue, and then processes directories one-by-one.
    """
    self.dirqueue = [os.path.abspath(firstdir)]
    while self.dirqueue:
      self.cwd = self.dirqueue.pop(0)
      logger.debug("processing directory %s", self.cwd)
      self.parse_file(os.path.join(self.cwd, "CTestTestfile.cmake"))

  def get_json(self):
    """Return a JSON representation of the test specification."""
    return json.dumps(
        [spec.as_odict() for _, spec in sorted(self.tests.items())],
        indent=2)

  def get_xml(self):
    """Return an XML representation of the test specification."""
    root = ET.Element("ctest")
    for _, spec in sorted(self.tests.items()):
      root.append(spec.as_element())
    buf = io.BytesIO()
    ET.ElementTree(root).write(buf)
    return buf.getvalue().decode("utf-8")


def setup_argparse(argparser):
  argparser.add_argument(
      "--log-level", default="info",
      choices=["debug", "info", "warning", "error"])
  mgroup = argparser.add_mutually_exclusive_group()
  mgroup.add_argument(
      "--json", action="store_const", dest="out_type", const="json")
  mgroup.add_argument(
      "--xml", action="store_const", dest="out_type", const="xml")
  argparser.add_argument("directory", nargs="?", default=".")


def main():
  logging.basicConfig(level=logging.INFO, format="%(levelname)s %(message)s")
  argparser = argparse.ArgumentParser(prog="ctest-to", description=__doc__)
  setup_argparse(argparser)
  args = argparser.parse_args()
  logger.setLevel(getattr(logging, args.log_level.upper()))
  ctx = ParseContext()
  ctx.start(args.directory)

  out_type = getattr(args, "out_type", "json")
  if out_type == "json":
    sys.stdout.write(ctx.get_json())
  elif out_type == "xml":
    xml_content = ctx.get_xml()
    try:
      import lxml.etree as etree
      sys.stdout.write(
          etree.tostring(etree.fromstring(xml_content), pretty_print=True)
          .decode("utf-8"))
    except ImportError:
      logger.warning("No lxml, cannot pretty-print")
      sys.stdout.write(xml_content)
  elif out_type is None:
    logger.info("No output requested")


if __name__ == "__main__":
  main()



# pylint: disable=too-many-lines
"""
Statement parser functions
"""

import importlib
import logging

from cmakelang import lex
from cmakelang.parse.additional_nodes import ShellCommandNode
from cmakelang.parse.argument_nodes import (
    ConditionalGroupNode, StandardParser, StandardParser2
)
from cmakelang.parse.common import NodeType, KwargBreaker, TreeNode
from cmakelang.parse.util import (
    IMPLICIT_PARG_TYPES,
    WHITESPACE_TOKENS,
    get_first_semantic_token,
    get_normalized_kwarg,
    get_tag,
    should_break,
)
from cmakelang.parse.funs import standard_funs

logger = logging.getLogger(__name__)


def get_funtree(cmdspec):
  kwargs = {}

  for kwarg, subspec in cmdspec.kwargs.items():
    if kwarg in ("if", "elseif", "while"):
      subparser = ConditionalGroupNode.parse
    elif kwarg == "COMMAND":
      subparser = ShellCommandNode.parse
    elif isinstance(subspec, (standard_funs.CommandSpec)):
      subparser = StandardParser2(subspec, get_funtree(subspec))
    else:
      raise ValueError("Unexpected kwarg spec of type {}"
                       .format(type(subspec)))

    kwargs[kwarg] = subparser

  return kwargs


SUBMODULE_NAMES = [
    "add_executable",
    "add_library",
    "add_xxx",
    "deprecated",
    "break",
    "external_project",
    "fetch_content",
    "foreach",
    "file",
    "install",
    "list",
    "miscellaneous",
    "random",
    "set",
    "set_target_properties",
]


def get_parse_db():
  """
  Returns a dictionary mapping statement name to parse functor for that
  statement.
  """

  parse_db = {}

  for subname in SUBMODULE_NAMES:
    submodule = importlib.import_module("cmakelang.parse.funs." + subname)
    submodule.populate_db(parse_db)

  for key in (
      "if", "else", "elseif", "endif", "while", "endwhile"):
    parse_db[key] = ConditionalGroupNode.parse

  for key in ("function", "macro"):
    parse_db[key] = StandardParser("1+")

  for key in ("endfunction", "endmacro"):
    parse_db[key] = StandardParser("?")

  parse_db.update(get_funtree(standard_funs.get_fn_spec()))
  return parse_db



import unittest

# pylint: disable=unused-import
from cmakelang.command_tests import (
    TestAddCustomCommand,
    TestConditional,
    TestComment,
    TestCustomCommand,
    TestExport,
    TestExternalProject,
    TestFile,
    TestForeach,
    TestInstall,
    TestSetTargetProperties,
    TestSet)

from cmakelang.command_tests.add_executable_tests \
    import TestAddExecutableCommand
from cmakelang.command_tests.add_library_tests \
    import TestAddLibraryCommand
from cmakelang.command_tests.misc_tests \
    import TestMiscFormatting


def main():
  unittest.main()


if __name__ == "__main__":
  main()



# pylint: disable=invalid-name
# How wide to allow formatted cmake files
line_width = 80

# How many spaces to tab for indent
tab_size = 2

# If true, separate control flow names from the parentheses with a space
separate_ctrl_name_with_space = False

# If true, separate function names from the parenthesis with a space
separate_fn_name_with_space = False



# -*- coding: utf-8 -*-
# PYTHON_ARGCOMPLETE_OK
"""
Parse cmake listfiles and format them nicely.

Formatting is configurable by providing a configuration file. The configuration
file can be in json, yaml, or python format. If no configuration file is
specified on the command line, cmake-format will attempt to find a suitable
configuration for each ``inputpath`` by checking recursively checking it's
parent directory up to the root of the filesystem. It will return the first
file it finds with a filename that matches '\\.?cmake-format(.yaml|.json|.py)'.

cmake-format can spit out the default configuration for you as starting point
for customization. Run with `--dump-config [yaml|json|python]`.
"""
from __future__ import unicode_literals

import argparse
import collections
try:
  from collections.abc import Mapping
except ImportError:
  from collections import Mapping
import io
import json
import logging
import os
import shutil
import sys

import cmakelang
from cmakelang import common
from cmakelang import configuration
from cmakelang import config_util
from cmakelang.format import formatter
from cmakelang import lex
from cmakelang import markup
from cmakelang import parse
from cmakelang.parse.argument_nodes import StandardParser2
from cmakelang.parse.common import NodeType, TreeNode
from cmakelang.parse.printer import dump_tree as dump_parse
from cmakelang.parse.funs import standard_funs


logger = logging.getLogger(__name__)


def detect_line_endings(infile_content):
  windows_count = infile_content.count('\r\n')
  unix_count = infile_content.count('\n') - windows_count
  if windows_count > unix_count:
    return 'windows'

  return 'unix'


def dump_markup(nodes, config, outfile=None, indent=None):
  """
  Print a tree of node objects for debugging purposes. Takes as input a full
  parse tree.
  """

  if indent is None:
    indent = ''

  if outfile is None:
    outfile = sys.stdout

  for idx, node in enumerate(nodes):
    if not isinstance(node, TreeNode):
      continue

    # outfile.write(indent)
    # if idx + 1 == len(nodes):
    #   outfile.write('└─ ')
    # else:
    #   outfile.write('├─ ')

    noderep = repr(node)
    if sys.version_info[0] < 3:
      noderep = getattr(noderep, 'decode')('utf-8')

    if node.node_type is NodeType.COMMENT:
      outfile.write(noderep)
      outfile.write('\n')
      inlines = []
      for token in node.children:
        assert isinstance(token, lex.Token)
        if token.type == lex.TokenType.COMMENT:
          inlines.append(token.spelling.strip().lstrip('#'))
      items = markup.parse(inlines, config)
      for item in items:
        outfile.write("{}\n".format(item))
      outfile.write("\n")

    if not hasattr(node, 'children'):
      continue

    if idx + 1 == len(nodes):
      dump_markup(node.children, config, outfile, indent + '    ')
    else:
      dump_markup(node.children, config, outfile, indent + '│   ')


def dump_parsedb(parsedb, outfile=None, indent=None):
  """
  Dump the parse database to a file
  """

  if indent is None:
    indent = ''

  if outfile is None:
    outfile = sys.stdout

  items = list(sorted(parsedb.items()))
  for idx, (name, value) in enumerate(items):
    outfile.write(indent)
    if idx + 1 == len(items):
      outfile.write('└─ ')
      subindent = indent + "   "
    else:
      outfile.write('├─ ')
      subindent = indent + "|  "

    outfile.write(name)
    if isinstance(value, StandardParser2):
      outfile.write(": {}\n".format(repr(value.cmdspec.pargs)))
      dump_parsedb(value.funtree, outfile, subindent)
    else:
      outfile.write(": {}\n".format(type(value)))


def process_file(config, infile_content, dump=None):
  """
  Parse the input cmake file, re-format it, and print to the output file.
  """

  outfile = io.StringIO(newline='')
  if config.format.line_ending == 'auto':
    detected = detect_line_endings(infile_content)
    config = config.clone()
    config.format.set_line_ending(detected)
  tokens = lex.tokenize(infile_content)
  if dump == "lex":
    for token in tokens:
      outfile.write("{}\n".format(token))
    return outfile.getvalue(), True
  first_token = lex.get_first_non_whitespace_token(tokens)
  parse_db = parse.funs.get_parse_db()
  parse_db.update(parse.funs.get_funtree(config.parse.fn_spec))

  if dump == "parsedb":
    dump_parsedb(parse_db, outfile)
    return outfile.getvalue(), True

  ctx = parse.ParseContext(parse_db, config=config)
  parse_tree = parse.parse(tokens, ctx)
  if dump == "parse":
    dump_parse([parse_tree], outfile)
    return outfile.getvalue(), True
  if dump == "markup":
    dump_markup([parse_tree], config, outfile)
    return outfile.getvalue(), True

  box_tree = formatter.layout_tree(parse_tree, config, first_token=first_token)
  if dump == "layout":
    formatter.dump_tree([box_tree], outfile)
    return outfile.getvalue(), True

  outstr = formatter.write_tree(box_tree, config, infile_content)
  if config.encode.emit_byteorder_mark:
    outstr = "\ufeff" + outstr

  outstr = formatter.replace_with_tabs(outstr, config)
  return (outstr, box_tree.reflow_valid)


def find_config_file(infile_path):
  """
  Search parent directories of an infile path and find a config file if
  one exists.
  """
  realpath = os.path.realpath(infile_path)
  if os.path.isdir(infile_path):
    head = infile_path
  else:
    head, _ = os.path.split(realpath)

  while head:
    for filename in ['.cmake-format',
                     '.cmake-format.py',
                     '.cmake-format.json',
                     '.cmake-format.yaml',
                     'cmake-format.py',
                     'cmake-format.json',
                     'cmake-format.yaml', ]:
      configpath = os.path.join(head, filename)
      if os.path.exists(configpath):
        return configpath
    head2, _ = os.path.split(head)
    if head == head2:
      break
    head = head2
  return None


def load_yaml(config_file):
  """
  Attempt to load yaml configuration from an opened file
  """
  import yaml
  try:
    from yaml import CLoader as Loader
  except ImportError:
    from yaml import Loader
  out = yaml.load(config_file, Loader=Loader)
  if out is None:
    return {}
  return out


def exec_pyconfig(configfile_path):
  _global = config_util.ExecGlobal(configfile_path)
  with io.open(configfile_path, 'r', encoding='utf-8') as infile:
    # pylint: disable=exec-used
    exec(infile.read(), _global)
  _global.pop("__builtins__", None)
  return _global


def try_get_configdict(configfile_path):
  """
  Try to read the configuration as yaml first, then json, then python.

  TODO(josh): maybe read the first line and look for some kind of comment
  sentinel to indicate the file format?
  """
  try:
    with io.open(configfile_path, 'r', encoding='utf-8') as config_file:
      return json.load(config_file)
  except:  # pylint: disable=bare-except
    pass

  try:
    with io.open(configfile_path, 'r', encoding='utf-8') as config_file:
      return load_yaml(config_file)
  except:  # pylint: disable=bare-except
    pass

  try:
    return exec_pyconfig(configfile_path)
  except:  # pylint: disable=bare-except
    pass

  raise RuntimeError("Failed to parse {} as any of yaml, json, or python"
                     .format(configfile_path))


def get_one_config_dict(configfile_path):
  """
  Return a dictionary of configuration options read from the given file path.
  If the filepath has a known extension then we parse it according to that
  extension. Otherwise we try to parse is using each parser one by one.
  """
  if not os.path.exists(configfile_path):
    raise common.UserError(
        "Desired config file does not exist: {}".format(configfile_path))

  if configfile_path.endswith('.json'):
    # NOTE(josh): an empty file is not valid JSON, but we don't need to be as
    # pedantic
    if os.stat(configfile_path).st_size == 0:
      return {}
    with io.open(configfile_path, 'r', encoding='utf-8') as config_file:
      try:
        return json.load(config_file)
      except ValueError as ex:
        message = (
            "Failed to parse json config file {}: {}"
            .format(configfile_path, ex))
        raise common.UserError(message)

  if configfile_path.endswith('.yaml'):
    with io.open(configfile_path, 'r', encoding='utf-8') as config_file:
      try:
        return load_yaml(config_file)
      except ValueError as ex:
        message = (
            "Failed to parse yaml config file {}: {}"
            .format(configfile_path, ex))
        raise common.UserError(message)

  if configfile_path.endswith('.py'):
    try:
      return exec_pyconfig(configfile_path)
    except Exception as ex:
      message = (
          "Failed to parse python config file {}: {}"
          .format(configfile_path, ex))
      raise common.UserError(message)

  return try_get_configdict(configfile_path)


def get_configdict(configfile_paths):
  include_queue = list(configfile_paths)
  config_dict = {}
  while include_queue:
    configfile_path = include_queue.pop(0)
    configfile_path = os.path.expanduser(configfile_path)
    increment_dict = get_one_config_dict(configfile_path)

    for include_path in increment_dict.pop("include", []):
      if not os.path.isabs(include_path):
        include_path = os.path.join(
            os.path.dirname(configfile_path), include_path)
      include_queue.append(include_path)
    map_merge(config_dict, increment_dict)
  return config_dict


def map_merge(output_map, increment_map):
  """
  Merge `increment_map` into `output_map` recursively.
  """
  for key, increment_value in increment_map.items():
    if key not in output_map:
      output_map[key] = increment_value
      continue

    existing_value = output_map[key]
    if isinstance(existing_value, Mapping):
      if isinstance(increment_value, Mapping):
        map_merge(existing_value, increment_value)
      else:
        logger.warning(
            "Cannot merge config %s of type %s into a dictionary",
            key, type(increment_value))
      continue

    output_map[key] = increment_value

  return output_map


def get_config(infile_path, configfile_paths):
  """
  If configfile_path is not none, then load the configuration. Otherwise search
  for a config file in the ancestry of the filesystem of infile_path and find
  a config file to load.
  """
  if configfile_paths is None:
    inferred_configpath = find_config_file(infile_path)
    if inferred_configpath is None:
      return {}
    configfile_paths = [inferred_configpath]

  return get_configdict(configfile_paths)


def yaml_odict_handler(dumper, value):
  """
  Represent ordered dictionaries as yaml maps.
  """
  return dumper.represent_mapping(u'tag:yaml.org,2002:map', value)


def yaml_register_odict(dumper):
  """
  Register an order dictionary handler with the given yaml dumper
  """
  dumper.add_representer(collections.OrderedDict, yaml_odict_handler)


def dump_config(args, config_dict, outfile):
  """
  Dump the default configuration to stdout
  """

  outfmt = args.dump_config
  config_dict.update(get_argdict(args))
  cfg = configuration.Configuration(**config_dict)
  # Don't dump default per-command configs
  for key in standard_funs.get_default_config():
    cfg.misc.per_command.pop(key, None)

  if outfmt == 'yaml':
    import yaml
    yaml_register_odict(yaml.SafeDumper)
    yaml_register_odict(yaml.Dumper)
    yaml.dump(cfg.as_odict(args.with_help, args.with_defaults),
              outfile, indent=2,
              default_flow_style=False, sort_keys=False)
    return
  if outfmt == 'json':
    json.dump(cfg.as_odict(args.with_help, args.with_defaults),
              outfile, indent=2)
    outfile.write('\n')
    return

  cfg.dump(outfile, with_help=args.with_help, with_defaults=args.with_defaults)


USAGE_STRING = """
cmake-format [-h]
             [--dump-config {yaml,json,python} | -i | -o OUTFILE_PATH]
             [-c CONFIG_FILE]
             infilepath [infilepath ...]
"""


class ExtendAction(argparse.Action):
  def __call__(self, parser, namespace, values, option_string=None):
    items = getattr(namespace, self.dest) or []
    items.extend(values)
    setattr(namespace, self.dest, items)


def setup_argparser(argparser):
  """
  Add argparse options to the parser.
  """
  argparser.register('action', 'extend', ExtendAction)
  argparser.add_argument('-v', '--version', action='version',
                         version=cmakelang.__version__)
  argparser.add_argument(
      '-l', '--log-level', default="info",
      choices=["error", "warning", "info", "debug"])

  mutex = argparser.add_mutually_exclusive_group()
  mutex.add_argument('--dump-config', choices=['yaml', 'json', 'python'],
                     default=None, const='python', nargs='?',
                     help='If specified, print the default configuration to '
                          'stdout and exit')
  mutex.add_argument(
      '--dump', choices=['lex', 'parse', 'parsedb', 'layout', 'markup'],
      default=None)

  argparser.add_argument(
      "--no-help", action="store_false", dest="with_help",
      help="When used with --dump-config, will omit helptext comments in the"
           " output"
  )
  argparser.add_argument(
      "--no-default", action="store_false", dest="with_defaults",
      help="When used with --dump-config, will omit any unmodified "
           "configuration value."
  )

  mutex = argparser.add_mutually_exclusive_group()
  mutex.add_argument('-i', '--in-place', action='store_true')
  mutex.add_argument(
      '--check', action='store_true',
      help="Exit with status code 0 if formatting would not change file "
           "contents, or status code 1 if it would")
  mutex.add_argument('-o', '--outfile-path', default=None,
                     help='Where to write the formatted file. '
                          'Default is stdout.')

  argparser.add_argument(
      '-c', '--config-files', nargs='+', action='extend',
      help='path to configuration file(s)')
  argparser.add_argument('infilepaths', nargs='*')

  configuration.Configuration().add_to_argparser(argparser)


def get_argdict(args):
  """Return a dictionary representation of the argparser `namespace` object
     returned from parse_args(). The returned dictionary will be suitable
     as a configuration kwargs dict. Any command line options that aren't
     configuration options are removed."""
  out = {}
  for key, value in vars(args).items():
    if key.startswith("_"):
      continue
    if hasattr(configuration.Configuration, key):
      continue
    # Remove common command line arguments
    if key in ["log_level", "outfile_path", "infilepaths", "config_files"]:
      continue
    # Remove --dump-config command line arguments
    if key in ["dump_config", "with_help", "with_defaults"]:
      continue
    # Remove cmake-format command line arguments
    if key in ["dump", "check", "in_place"]:
      continue
    # Remove cmake-lint command line arguments
    if key in ["suppress_decorations"]:
      continue
    if value is None:
      continue
    out[key] = value

  return out


def onefile_main(infile_path, args, argparse_dict):
  """
  Find config, open file, process, write result
  """
  # NOTE(josh): have to load config once for every file, because we may pick
  # up a new config file location for each path
  if infile_path == '-':
    config_dict = get_config(os.getcwd(), args.config_files)
  else:
    config_dict = get_config(infile_path, args.config_files)

  cfg = configuration.Configuration(**config_dict)
  cfg.legacy_consume(argparse_dict)

  if cfg.format.disable:
    return

  if infile_path == '-':
    infile = io.open(os.dup(sys.stdin.fileno()),
                     mode='r', encoding=cfg.encode.input_encoding, newline='')
  else:
    infile = io.open(
        infile_path, 'r', encoding=cfg.encode.input_encoding, newline='')
  with infile:
    intext = infile.read()

  try:
    outtext, reflow_valid = process_file(cfg, intext, args.dump)
    if cfg.format.require_valid_layout and not reflow_valid:
      raise common.FormatError("Failed to format {}".format(infile_path))
  except:
    logger.warning('While processing %s', infile_path)
    raise

  if args.check:
    if intext != outtext:
      raise common.FormatError("Check failed: {}".format(infile_path))
    return

  if args.in_place:
    if intext == outtext:
      logger.debug("No delta for %s", infile_path)
      return
    tempfile_path = infile_path + ".cmf-temp"
    outfile = io.open(
        tempfile_path, 'w', encoding=cfg.encode.output_encoding, newline='')
  else:
    if args.outfile_path == '-':
      # NOTE(josh): The behavior of sys.stdout is different in python2 and
      # python3. sys.stdout is opened in 'w' mode which means that write()
      # takes strings in python2 and python3 and, in particular, in python3
      # it does not take byte arrays. io.StreamWriter will write to
      # it with byte arrays (assuming it was opened with 'wb'). So we use
      # io.open instead of open in this case
      outfile = io.open(
          os.dup(sys.stdout.fileno()),
          mode='w', encoding=cfg.encode.output_encoding, newline='')
    else:
      outfile = io.open(
          args.outfile_path, 'w', encoding=cfg.encode.output_encoding,
          newline='')

  with outfile:
    outfile.write(outtext)

  if args.in_place:
    shutil.copymode(infile_path, tempfile_path)
    shutil.move(tempfile_path, infile_path)


def inner_main():
  """Parse arguments, open files, start work."""

  arg_parser = argparse.ArgumentParser(
      description=__doc__,
      formatter_class=argparse.RawDescriptionHelpFormatter,
      usage=USAGE_STRING)

  setup_argparser(arg_parser)
  try:
    import argcomplete
    argcomplete.autocomplete(arg_parser)
  except ImportError:
    pass
  args = arg_parser.parse_args()
  logging.getLogger().setLevel(getattr(logging, args.log_level.upper()))

  if args.dump_config:
    config_dict = get_config(os.getcwd(), args.config_files)
    dump_config(args, config_dict, sys.stdout)
    sys.exit(0)

  if args.dump_config or args.dump:
    assert not args.in_place, \
        ("-i/--in-place not allowed when dumping")

  assert args.in_place is False or args.outfile_path is None, \
      "if inplace is specified than outfile is invalid"
  assert (len(args.infilepaths) == 1
          or (args.in_place is True or args.outfile_path is None)), \
      ("if more than one input file is specified, then formatting must be done"
       " in-place or written to stdout")

  if args.outfile_path is None:
    args.outfile_path = '-'

  if '-' in args.infilepaths:
    assert len(args.infilepaths) == 1, \
        "You cannot mix stdin as an input with other input files"
    assert args.outfile_path == '-', \
        "If stdin is the input file, then stdout must be the output file"

  argparse_dict = get_argdict(args)

  returncode = 0
  for infile_path in args.infilepaths:
    try:
      onefile_main(infile_path, args, argparse_dict)
    except common.FormatError as ex:
      logger.error(ex.msg)
      returncode = 1

  return returncode


def main():
  # set up main logger, which logs everything. We'll leave this one logging
  # to the console
  format_str = '%(levelname)-4s %(filename)s:%(lineno)-3s: %(message)s'
  logging.basicConfig(level=logging.INFO,
                      format=format_str,
                      filemode='w')

  try:
    return inner_main()
  except common.UserError as ex:
    logger.fatal(ex.msg)
    return 1
  except common.InternalError as ex:
    logger.exception(ex.msg)
    return 1
  except AssertionError as ex:
    logger.exception(
        "An internal error occured. Please consider filing a bug report at "
        "github.com/cheshirekow/cmakelang/issues")
    return 1


if __name__ == '__main__':
  sys.exit(main())



"""
Given a staging directory and a file containing a list of what should be in
that directory, delete any files that shouldn't be there.
"""

import argparse
import io
import logging
import os
import sys

logger = logging.getLogger(__name__)


def setup_argparser(argparser):
  argparser.add_argument(
      "manifest_path", help="path to the manifest file")
  argparser.add_argument(
      "stage_path", help="path to the staging directory")


def get_argdict(args):
  out = {}
  for key, value in vars(args).items():
    if key.startswith("_"):
      continue
    out[key] = value
  return out


def inner_main(manifest_path, stage_path):
  with io.open(manifest_path, "r", encoding="utf-8") as infile:
    manifest = set(line.strip() for line in infile)

  for directory, _dirnames, filenames in os.walk(stage_path):
    relpath_dir = os.path.relpath(directory, stage_path)

    for filename in filenames:
      if relpath_dir == ".":
        relpath_file = filename
      else:
        relpath_file = os.path.join(relpath_dir, filename)
      fullpath_file = os.path.join(directory, filename)

      if relpath_file not in manifest and fullpath_file not in manifest:
        logger.info("Deleting %s", relpath_file)
        os.unlink(fullpath_file)


def main():
  logging.basicConfig(level=logging.INFO)
  argparser = argparse.ArgumentParser(description=__doc__)
  setup_argparser(argparser)
  try:
    import argcomplete
    argcomplete.autocomplete(argparser)
  except ImportError:
    pass
  args = argparser.parse_args()
  inner_main(**get_argdict(args))
  return 0


if __name__ == "__main__":
  sys.exit(main())



import logging

from cmakelang.lex import TokenType
from cmakelang.parse.argument_nodes import (
    StandardArgTree,
    StandardParser)
from cmakelang.parse.util import get_first_semantic_token

logger = logging.getLogger(__name__)


def null_parser(ctx, tokens, breakstack):  # pylint: disable=unused-argument
  return None


SORT_KWARGS = {
    "COMPARE": StandardParser(1, flags=["STRING", "FILE_BASENAME"]),
    "CASE": StandardParser(1, flags=["SENSITIVE", "INSENSITIVE"]),
    "ORDER": StandardParser(1, flags=["ASCENDING", "DESCENDING"])
}

TRANSFORM_KWARGS = {
    # actions
    "APPEND": StandardParser(1),
    "PREPEND": StandardParser(1),
    "TOLOWER": null_parser,
    "TOUPPER": null_parser,
    "STRIP": null_parser,
    "GENEX_STRIP": null_parser,
    "REPLACE": StandardParser(2),
    # selectors
    "AT": StandardParser("+"),
    "FOR": StandardParser("2+"),
    "REGEX": StandardParser(1)
}


def parse_list(ctx, tokens, breakstack):
  """
  Reading
    list(LENGTH <list> <out-var>)
    list(GET <list> <element index> [<index> ...] <out-var>)
    list(JOIN <list> <glue> <out-var>)
    list(SUBLIST <list> <begin> <length> <out-var>)

  Search
    list(FIND <list> <value> <out-var>)

  Modification
    list(APPEND <list> [<element>...])
    list(FILTER <list> {INCLUDE | EXCLUDE} REGEX <regex>)
    list(INSERT <list> <index> [<element>...])
    list(POP_BACK <list> [<out-var>...])
    list(POP_FRONT <list> [<out-var>...])
    list(PREPEND <list> [<element>...])
    list(REMOVE_ITEM <list> <value>...)
    list(REMOVE_AT <list> <index>...)
    list(REMOVE_DUPLICATES <list>)
    list(TRANSFORM <list> <ACTION> [...])

  Ordering
    list(REVERSE <list>)
    list(SORT <list> [...])

  TRANSFORM actions:
    list(TRANSFORM <list> <APPEND|PREPEND> <value> ...)
    list(TRANSFORM <list> <TOLOWER|TOUPPER> ...)
    list(TRANSFORM <list> STRIP ...)
    list(TRANSFORM <list> GENEX_STRIP ...)
    list(TRANSFORM <list> REPLACE <regular_expression>
                                  <replace_expression> ...)

  TRANSFORM selectors
    list(TRANSFORM <list> <ACTION> AT <index> [<index> ...] ...)
    list(TRANSFORM <list> <ACTION> FOR <start> <stop> [<step>] ...)
    list(TRANSFORM <list> <ACTION> REGEX <regular_expression> ...)

  :see: https://cmake.org/cmake/help/latest/command/list.html
  """

  descriminator_token = get_first_semantic_token(tokens)
  if (descriminator_token is None or
      descriminator_token.type is TokenType.RIGHT_PAREN):
    location = ()
    if tokens:
      location = tokens[0].get_location()
    logger.warning("Invalid empty list() command at %s", location)
    ctx.lint_ctx.record_lint("E1120", location=location)
    return StandardArgTree.parse(ctx, tokens, npargs='*', kwargs={}, flags=[],
                                 breakstack=breakstack)

  if descriminator_token.type is TokenType.DEREF:
    ctx.lint_ctx.record_lint(
        "C0114", location=descriminator_token.get_location())
    return StandardArgTree.parse(ctx, tokens, npargs='*', kwargs={}, flags=[],
                                 breakstack=breakstack)

  descriminator = descriminator_token.spelling.upper()
  parsemap = {
      "LENGTH": StandardParser(3, flags=["LENGTH"]),
      "GET": StandardParser("4+", flags=["GET"]),
      "JOIN": StandardParser(4, flags=["JOIN"]),
      "SUBLIST": StandardParser(5, flags=["SUBLIST"]),
      "FIND": StandardParser(4, flags=["FIND"]),
      "APPEND": StandardParser("2+", flags=["APPEND"]),
      "FILTER": StandardParser(
          5, flags=["FILTER", "INCLUDE", "EXCLUDE", "REGEX"]),
      "INSERT": StandardParser("4+", flags=["INSERT"]),
      "POP_BACK": StandardParser("2+", flags=["POP_BACK"]),
      "POP_FRONT": StandardParser("2+", flags=["POP_FRONT"]),
      "PREPEND": StandardParser("2+", flags=["PREPEND"]),
      "REMOVE_ITEM": StandardParser("3+", flags=["REMOVE_ITEM"]),
      "REMOVE_AT": StandardParser("3+", flags=["REMOVE_AT"]),
      "REMOVE_DUPLICATES": StandardParser(2, flags=["REMOVE_DUPLICATES"]),
      "TRANSFORM": StandardParser(
          2, flags=["TRANSFORM"], kwargs=TRANSFORM_KWARGS),
      "REVERSE": StandardParser(2, flags=["REVERSE"]),
      "SORT": StandardParser(2, flags=["SORT"], kwargs=SORT_KWARGS)
  }

  if descriminator not in parsemap:
    logger.warning(
        "Indeterminate form of file() command \"%s\" at %s", descriminator,
        tokens[0].location())
    ctx.lint_ctx.record_lint(
        "E1126", location=descriminator_token.get_location())
    return StandardArgTree.parse(ctx, tokens, npargs='*', kwargs={}, flags=[],
                                 breakstack=breakstack)

  return parsemap[descriminator](ctx, tokens, breakstack)


def populate_db(parse_db):
  parse_db["list"] = parse_list



# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import unicode_literals

import logging

from cmakelang import lex
from cmakelang.parse.util import (
    WHITESPACE_TOKENS, pargs_are_full, should_break, get_normalized_kwarg
)
from cmakelang.parse.common import (
    NodeType, TreeNode,
)
from cmakelang.parse.simple_nodes import CommentNode, OnOffNode
from cmakelang.parse.argument_nodes import (
    ArgGroupNode, PositionalGroupNode, PositionalParser, PositionalSpec,
    StandardArgTree)

logger = logging.getLogger(__name__)

# pylint: disable=W0221


class TupleGroupNode(ArgGroupNode):
  """A positinal argument group where each argument is a tuple of tokens."""

  @classmethod
  def parse(cls, ctx, tokens, npargs, ntup, flags, breakstack):
    """Parse a continuous sequence of `npargs` positional argument pairs.
    If npargs is an integer we will consume exactly that many arguments.
    If it is not an integer then it is a string meaning:

    * "?": zero or one
    * "*": zero or more
    * "+": one or more
    """

    tree = cls()
    tree.spec = PositionalSpec(ntup, flags=flags, legacy=True)
    subtree = None
    active_depth = tree

    npargs_consumed = 0
    ntup_consumed = 0

    while tokens:
      # Break if we have consumed enough positional arguments
      if pargs_are_full(npargs, npargs_consumed):
        break

      # Break if the next token belongs to a parent parser, i.e. if it
      # matches a keyword argument of something higher in the stack, or if
      # it closes a parent group.
      if should_break(tokens[0], breakstack):
        break

      # Otherwise we will consume the token
      token = tokens.pop(0)

      # If it is a whitespace token then put it directly in the parse tree at
      # the current depth
      if token.type in WHITESPACE_TOKENS:
        active_depth.children.append(token)
        continue

      # If it's a comment token not associated with an argument, then put it
      # directly into the parse tree at the current depth
      if token.type in (lex.TokenType.COMMENT,
                        lex.TokenType.BRACKET_COMMENT):
        child = CommentNode()
        tree.children.append(child)
        child.children.append(token)
        continue

      # If it's a sentinel comment, then add it at the current depth
      if tokens[0].type in (lex.TokenType.FORMAT_OFF,
                            lex.TokenType.FORMAT_ON):
        tree.children.append(OnOffNode.consume(ctx, tokens))
        continue

      if subtree is None:
        subtree = PositionalGroupNode()
        subtree.spec = PositionalSpec(2, False, [], flags)
        tree.children.append(subtree)
        ntup_consumed = 0

      # Otherwise is it is a positional argument, so add it to the tree as such
      if get_normalized_kwarg(token) in flags:
        child = TreeNode(NodeType.FLAG)
      else:
        child = TreeNode(NodeType.ARGUMENT)

      child.children.append(token)
      CommentNode.consume_trailing(ctx, tokens, child)
      subtree.children.append(child)
      ntup_consumed += 1

      if ntup_consumed >= ntup:
        npargs_consumed += 1
        subtree = None

    return tree


class TupleParser(object):
  def __init__(self, ntup, npargs=None, flags=None):
    if npargs is None:
      npargs = "*"
    if flags is None:
      flags = []

    self.npargs = npargs
    self.ntup = ntup
    self.flags = flags

  def __call__(self, ctx, tokens, breakstack):
    return TupleGroupNode.parse(
        ctx, tokens, self.npargs, self.ntup, self.flags, breakstack)


class ShellCommandNode(StandardArgTree):
  """Shell commands are children of a `COMMAND` keyword argument and are
     common enough to warrant their own node. We also will likely want some
     special formatting rules for these nodes.
  """

  @classmethod
  def parse(cls, ctx, tokens, breakstack):
    """
    Parser for the COMMAND kwarg lists in the form of::

        COMMAND foo --long-flag1 arg1 arg2 --long-flag2 -a -b -c arg3 arg4

    The parser acts very similar to a standard parser where `--xxx` is treated
    as a keyword argument and `-x` is treated as a flag.
    """
    tree = super(ShellCommandNode, cls).parse(
        ctx, tokens, '+', {}, [], breakstack)
    for pgroup in tree.parg_groups:
      pgroup.tags.append("cmdline")
    return tree


class PatternNode(StandardArgTree):
  """Patterns are children of a `PATTERN` keyword argument and are common
     enough to warrent their own node."""

  @classmethod
  def parse(cls, ctx, tokens, breakstack):
    """
    ::

      [PATTERN <pattern> | REGEX <regex>]
      [EXCLUDE] [PERMISSIONS <permissions>...]
    """
    return super(PatternNode, cls).parse(
        ctx, tokens,
        npargs='+',
        kwargs={"PERMISSIONS": PositionalParser('+'), },
        flags=["EXCLUDE"],
        breakstack=breakstack
    )


class FlagGroupNode(PositionalGroupNode):
  """A positinal group where each argument is a flag."""

  @classmethod
  def parse(cls, ctx, tokens, flags, breakstack):
    """
    Parse a continuous sequence of flags
    """

    tree = cls()
    tree.spec = PositionalSpec('+', False, [], flags)
    while tokens:
      # Break if the next token belongs to a parent parser, i.e. if it
      # matches a keyword argument of something higher in the stack, or if
      # it closes a parent group.
      if should_break(tokens[0], breakstack):
        break

      # If it is a whitespace token then put it directly in the parse tree at
      # the current depth
      if tokens[0].type in WHITESPACE_TOKENS:
        tree.children.append(tokens.pop(0))
        continue

      # Break if the next token is not a known flag
      if tokens[0].spelling.upper() not in flags:
        break

      # Otherwise is it is a flag, so add it to the tree as such
      child = TreeNode(NodeType.FLAG)
      child.children.append(tokens.pop(0))
      CommentNode.consume_trailing(ctx, tokens, child)
      tree.children.append(child)

    return tree




import unittest
import subprocess

from cmakelang.parse.funs import get_parse_db

IGNORE_LIST = [
    "endforeach",
    "endfunction",
    "endmacro"
]


class TestCommandDatabase(unittest.TestCase):
  """
  Execute cmake and ensure that all cmake commands are in the database
  """

  def test_all_commands_in_db(self):
    missing_commands = []
    proc = subprocess.Popen(
        ["cmake", "--help-command-list"],
        stdout=subprocess.PIPE)

    parse_db = get_parse_db()

    ignore = IGNORE_LIST
    with proc.stdout as infile:
      for line in infile:
        command = line.strip().decode("utf-8")
        if command not in parse_db and command not in ignore:
          missing_commands.append(command)
    proc.wait()

    message = "Missing commands:\n  " + "\n  ".join(sorted(missing_commands))
    self.assertFalse(bool(missing_commands), msg=message)


if __name__ == "__main__":
  unittest.main()



# pylint: skip-file
"""
Database of known property names
NOTE: this file is automatically generated by the script
python -Bm cmakelang.tools.parse_cmake_help properties
"""
import re


PATTERNS = [
    "(?P<CONFIG>.*)_OUTPUT_NAME",
    "(?P<CONFIG>.*)_POSTFIX",
    "(?P<LANG>.*)_CLANG_TIDY",
    "(?P<LANG>.*)_COMPILER_LAUNCHER",
    "(?P<LANG>.*)_CPPCHECK",
    "(?P<LANG>.*)_CPPLINT",
    "(?P<LANG>.*)_INCLUDE_WHAT_YOU_USE",
    "(?P<LANG>.*)_VISIBILITY_PRESET",
    "ABSTRACT",
    "ADDITIONAL_MAKE_CLEAN_FILES",
    "ADVANCED",
    "ALIASED_TARGET",
    "ALLOW_DUPLICATE_CUSTOM_TARGETS",
    "ANDROID_ANT_ADDITIONAL_OPTIONS",
    "ANDROID_API",
    "ANDROID_API_MIN",
    "ANDROID_ARCH",
    "ANDROID_ASSETS_DIRECTORIES",
    "ANDROID_GUI",
    "ANDROID_JAR_DEPENDENCIES",
    "ANDROID_JAR_DIRECTORIES",
    "ANDROID_JAVA_SOURCE_DIR",
    "ANDROID_NATIVE_LIB_DEPENDENCIES",
    "ANDROID_NATIVE_LIB_DIRECTORIES",
    "ANDROID_PROCESS_MAX",
    "ANDROID_PROGUARD",
    "ANDROID_PROGUARD_CONFIG_PATH",
    "ANDROID_SECURE_PROPS_PATH",
    "ANDROID_SKIP_ANT_STEP",
    "ANDROID_STL_TYPE",
    "ARCHIVE_OUTPUT_DIRECTORY",
    "ARCHIVE_OUTPUT_DIRECTORY_(?P<CONFIG>.*)",
    "ARCHIVE_OUTPUT_NAME",
    "ARCHIVE_OUTPUT_NAME_(?P<CONFIG>.*)",
    "ATTACHED_FILES",
    "ATTACHED_FILES_ON_FAIL",
    "AUTOGEN_BUILD_DIR",
    "AUTOGEN_SOURCE_GROUP",
    "AUTOGEN_TARGETS_FOLDER",
    "AUTOGEN_TARGET_DEPENDS",
    "AUTOMOC",
    "AUTOMOC_COMPILER_PREDEFINES",
    "AUTOMOC_DEPEND_FILTERS",
    "AUTOMOC_MACRO_NAMES",
    "AUTOMOC_MOC_OPTIONS",
    "AUTOMOC_SOURCE_GROUP",
    "AUTOMOC_TARGETS_FOLDER",
    "AUTORCC",
    "AUTORCC_OPTIONS",
    "AUTORCC_OPTIONS",
    "AUTORCC_SOURCE_GROUP",
    "AUTOUIC",
    "AUTOUIC_OPTIONS",
    "AUTOUIC_OPTIONS",
    "AUTOUIC_SEARCH_PATHS",
    "BINARY_DIR",
    "BINARY_DIR",
    "BUILDSYSTEM_TARGETS",
    "BUILD_RPATH",
    "BUILD_WITH_INSTALL_NAME_DIR",
    "BUILD_WITH_INSTALL_RPATH",
    "BUNDLE",
    "BUNDLE_EXTENSION",
    "CACHE_VARIABLES",
    "CLEAN_NO_CUSTOM",
    "CMAKE_CONFIGURE_DEPENDS",
    "CMAKE_CXX_KNOWN_FEATURES",
    "CMAKE_C_KNOWN_FEATURES",
    "COMPATIBLE_INTERFACE_BOOL",
    "COMPATIBLE_INTERFACE_NUMBER_MAX",
    "COMPATIBLE_INTERFACE_NUMBER_MIN",
    "COMPATIBLE_INTERFACE_STRING",
    "COMPILE_DEFINITIONS",
    "COMPILE_DEFINITIONS",
    "COMPILE_DEFINITIONS",
    "COMPILE_DEFINITIONS_(?P<CONFIG>.*)",
    "COMPILE_DEFINITIONS_(?P<CONFIG>.*)",
    "COMPILE_DEFINITIONS_(?P<CONFIG>.*)",
    "COMPILE_FEATURES",
    "COMPILE_FLAGS",
    "COMPILE_FLAGS",
    "COMPILE_OPTIONS",
    "COMPILE_OPTIONS",
    "COMPILE_PDB_NAME",
    "COMPILE_PDB_NAME_(?P<CONFIG>.*)",
    "COMPILE_PDB_OUTPUT_DIRECTORY",
    "COMPILE_PDB_OUTPUT_DIRECTORY_(?P<CONFIG>.*)",
    "COST",
    "CPACK_DESKTOP_SHORTCUTS",
    "CPACK_NEVER_OVERWRITE",
    "CPACK_PERMANENT",
    "CPACK_STARTUP_SHORTCUTS",
    "CPACK_START_MENU_SHORTCUTS",
    "CPACK_WIX_ACL",
    "CROSSCOMPILING_EMULATOR",
    "CUDA_EXTENSIONS",
    "CUDA_PTX_COMPILATION",
    "CUDA_RESOLVE_DEVICE_SYMBOLS",
    "CUDA_SEPARABLE_COMPILATION",
    "CUDA_STANDARD",
    "CUDA_STANDARD_REQUIRED",
    "CXX_EXTENSIONS",
    "CXX_STANDARD",
    "CXX_STANDARD_REQUIRED",
    "C_EXTENSIONS",
    "C_STANDARD",
    "C_STANDARD_REQUIRED",
    "DEBUG_CONFIGURATIONS",
    "DEBUG_POSTFIX",
    "DEFINE_SYMBOL",
    "DEFINITIONS",
    "DEPENDS",
    "DEPLOYMENT_REMOTE_DIRECTORY",
    "DISABLED",
    "DISABLED_FEATURES",
    "ECLIPSE_EXTRA_NATURES",
    "ENABLED_FEATURES",
    "ENABLED_LANGUAGES",
    "ENABLE_EXPORTS",
    "ENVIRONMENT",
    "EXCLUDE_FROM_ALL",
    "EXCLUDE_FROM_ALL",
    "EXCLUDE_FROM_DEFAULT_BUILD",
    "EXCLUDE_FROM_DEFAULT_BUILD_(?P<CONFIG>.*)",
    "EXPORT_NAME",
    "EXTERNAL_OBJECT",
    "EchoString",
    "FAIL_REGULAR_EXPRESSION",
    "FIND_LIBRARY_USE_LIB32_PATHS",
    "FIND_LIBRARY_USE_LIB64_PATHS",
    "FIND_LIBRARY_USE_LIBX32_PATHS",
    "FIND_LIBRARY_USE_OPENBSD_VERSIONING",
    "FIXTURES_CLEANUP",
    "FIXTURES_REQUIRED",
    "FIXTURES_SETUP",
    "FOLDER",
    "FRAMEWORK",
    "FRAMEWORK_VERSION",
    "Fortran_FORMAT",
    "Fortran_FORMAT",
    "Fortran_MODULE_DIRECTORY",
    "GENERATED",
    "GENERATOR_FILE_NAME",
    "GENERATOR_IS_MULTI_CONFIG",
    "GLOBAL_DEPENDS_DEBUG_MODE",
    "GLOBAL_DEPENDS_NO_CYCLES",
    "GNUtoMS",
    "HAS_CXX",
    "HEADER_FILE_ONLY",
    "HELPSTRING",
    "IMPLICIT_DEPENDS_INCLUDE_TRANSFORM",
    "IMPLICIT_DEPENDS_INCLUDE_TRANSFORM",
    "IMPORTED",
    "IMPORTED_CONFIGURATIONS",
    "IMPORTED_IMPLIB",
    "IMPORTED_IMPLIB_(?P<CONFIG>.*)",
    "IMPORTED_LIBNAME",
    "IMPORTED_LIBNAME_(?P<CONFIG>.*)",
    "IMPORTED_LINK_DEPENDENT_LIBRARIES",
    "IMPORTED_LINK_DEPENDENT_LIBRARIES_(?P<CONFIG>.*)",
    "IMPORTED_LINK_INTERFACE_LANGUAGES",
    "IMPORTED_LINK_INTERFACE_LANGUAGES_(?P<CONFIG>.*)",
    "IMPORTED_LINK_INTERFACE_LIBRARIES",
    "IMPORTED_LINK_INTERFACE_LIBRARIES_(?P<CONFIG>.*)",
    "IMPORTED_LINK_INTERFACE_MULTIPLICITY",
    "IMPORTED_LINK_INTERFACE_MULTIPLICITY_(?P<CONFIG>.*)",
    "IMPORTED_LOCATION",
    "IMPORTED_LOCATION_(?P<CONFIG>.*)",
    "IMPORTED_NO_SONAME",
    "IMPORTED_NO_SONAME_(?P<CONFIG>.*)",
    "IMPORTED_OBJECTS",
    "IMPORTED_OBJECTS_(?P<CONFIG>.*)",
    "IMPORTED_SONAME",
    "IMPORTED_SONAME_(?P<CONFIG>.*)",
    "IMPORT_PREFIX",
    "IMPORT_SUFFIX",
    "INCLUDE_DIRECTORIES",
    "INCLUDE_DIRECTORIES",
    "INCLUDE_REGULAR_EXPRESSION",
    "INSTALL_NAME_DIR",
    "INSTALL_RPATH",
    "INSTALL_RPATH_USE_LINK_PATH",
    "INTERFACE_AUTOUIC_OPTIONS",
    "INTERFACE_COMPILE_DEFINITIONS",
    "INTERFACE_COMPILE_FEATURES",
    "INTERFACE_COMPILE_OPTIONS",
    "INTERFACE_INCLUDE_DIRECTORIES",
    "INTERFACE_LINK_LIBRARIES",
    "INTERFACE_POSITION_INDEPENDENT_CODE",
    "INTERFACE_SOURCES",
    "INTERFACE_SYSTEM_INCLUDE_DIRECTORIES",
    "INTERPROCEDURAL_OPTIMIZATION",
    "INTERPROCEDURAL_OPTIMIZATION",
    "INTERPROCEDURAL_OPTIMIZATION_(?P<CONFIG>.*)",
    "INTERPROCEDURAL_OPTIMIZATION_(?P<CONFIG>.*)",
    "IN_TRY_COMPILE",
    "IOS_INSTALL_COMBINED",
    "JOB_POOLS",
    "JOB_POOL_COMPILE",
    "JOB_POOL_LINK",
    "KEEP_EXTENSION",
    "LABELS",
    "LABELS",
    "LABELS",
    "LABELS",
    "LANGUAGE",
    "LIBRARY_OUTPUT_DIRECTORY",
    "LIBRARY_OUTPUT_DIRECTORY_(?P<CONFIG>.*)",
    "LIBRARY_OUTPUT_NAME",
    "LIBRARY_OUTPUT_NAME_(?P<CONFIG>.*)",
    "LINKER_LANGUAGE",
    "LINK_DEPENDS",
    "LINK_DEPENDS_NO_SHARED",
    "LINK_DIRECTORIES",
    "LINK_FLAGS",
    "LINK_FLAGS_(?P<CONFIG>.*)",
    "LINK_INTERFACE_LIBRARIES",
    "LINK_INTERFACE_LIBRARIES_(?P<CONFIG>.*)",
    "LINK_INTERFACE_MULTIPLICITY",
    "LINK_INTERFACE_MULTIPLICITY_(?P<CONFIG>.*)",
    "LINK_LIBRARIES",
    "LINK_SEARCH_END_STATIC",
    "LINK_SEARCH_START_STATIC",
    "LINK_WHAT_YOU_USE",
    "LISTFILE_STACK",
    "LOCATION",
    "LOCATION",
    "LOCATION_(?P<CONFIG>.*)",
    "MACOSX_BUNDLE",
    "MACOSX_BUNDLE_INFO_PLIST",
    "MACOSX_FRAMEWORK_INFO_PLIST",
    "MACOSX_PACKAGE_LOCATION",
    "MACOSX_RPATH",
    "MACROS",
    "MANUALLY_ADDED_DEPENDENCIES",
    "MAP_IMPORTED_CONFIG_(?P<CONFIG>.*)",
    "MEASUREMENT",
    "MODIFIED",
    "NAME",
    "NO_SONAME",
    "NO_SYSTEM_FROM_IMPORTED",
    "OBJECT_DEPENDS",
    "OBJECT_OUTPUTS",
    "OSX_ARCHITECTURES",
    "OSX_ARCHITECTURES_(?P<CONFIG>.*)",
    "OUTPUT_NAME",
    "OUTPUT_NAME_(?P<CONFIG>.*)",
    "PACKAGES_FOUND",
    "PACKAGES_NOT_FOUND",
    "PARENT_DIRECTORY",
    "PASS_REGULAR_EXPRESSION",
    "PDB_NAME",
    "PDB_NAME_(?P<CONFIG>.*)",
    "PDB_OUTPUT_DIRECTORY",
    "PDB_OUTPUT_DIRECTORY_(?P<CONFIG>.*)",
    "POSITION_INDEPENDENT_CODE",
    "POST_INSTALL_SCRIPT",
    "PREDEFINED_TARGETS_FOLDER",
    "PREFIX",
    "PRE_INSTALL_SCRIPT",
    "PRIVATE_HEADER",
    "PROCESSORS",
    "PROJECT_LABEL",
    "PUBLIC_HEADER",
    "REPORT_UNDEFINED_PROPERTIES",
    "REQUIRED_FILES",
    "RESOURCE",
    "RESOURCE_LOCK",
    "RULE_LAUNCH_COMPILE",
    "RULE_LAUNCH_COMPILE",
    "RULE_LAUNCH_COMPILE",
    "RULE_LAUNCH_CUSTOM",
    "RULE_LAUNCH_CUSTOM",
    "RULE_LAUNCH_CUSTOM",
    "RULE_LAUNCH_LINK",
    "RULE_LAUNCH_LINK",
    "RULE_LAUNCH_LINK",
    "RULE_MESSAGES",
    "RUNTIME_OUTPUT_DIRECTORY",
    "RUNTIME_OUTPUT_DIRECTORY_(?P<CONFIG>.*)",
    "RUNTIME_OUTPUT_NAME",
    "RUNTIME_OUTPUT_NAME_(?P<CONFIG>.*)",
    "RUN_SERIAL",
    "SKIP_AUTOGEN",
    "SKIP_AUTOMOC",
    "SKIP_AUTORCC",
    "SKIP_AUTOUIC",
    "SKIP_BUILD_RPATH",
    "SKIP_RETURN_CODE",
    "SOURCES",
    "SOURCE_DIR",
    "SOURCE_DIR",
    "SOVERSION",
    "STATIC_LIBRARY_FLAGS",
    "STATIC_LIBRARY_FLAGS_(?P<CONFIG>.*)",
    "STRINGS",
    "SUBDIRECTORIES",
    "SUFFIX",
    "SYMBOLIC",
    "TARGET_ARCHIVES_MAY_BE_SHARED_LIBS",
    "TARGET_MESSAGES",
    "TARGET_SUPPORTS_SHARED_LIBS",
    "TEST_INCLUDE_FILE",
    "TEST_INCLUDE_FILES",
    "TIMEOUT",
    "TIMEOUT_AFTER_MATCH",
    "TYPE",
    "TYPE",
    "USE_FOLDERS",
    "VALUE",
    "VARIABLES",
    "VERSION",
    "VISIBILITY_INLINES_HIDDEN",
    "VS_CONFIGURATION_TYPE",
    "VS_COPY_TO_OUT_DIR",
    "VS_CSHARP_(?P<tagname>.*)",
    "VS_DEBUGGER_WORKING_DIRECTORY",
    "VS_DEPLOYMENT_CONTENT",
    "VS_DEPLOYMENT_LOCATION",
    "VS_DESKTOP_EXTENSIONS_VERSION",
    "VS_DOTNET_REFERENCEPROP_(?P<refname>.*)_TAG_(?P<tagname>.*)",
    "VS_DOTNET_REFERENCES",
    "VS_DOTNET_REFERENCES_COPY_LOCAL",
    "VS_DOTNET_REFERENCE_(?P<refname>.*)",
    "VS_DOTNET_TARGET_FRAMEWORK_VERSION",
    "VS_GLOBAL_(?P<variable>.*)",
    "VS_GLOBAL_KEYWORD",
    "VS_GLOBAL_PROJECT_TYPES",
    "VS_GLOBAL_ROOTNAMESPACE",
    "VS_GLOBAL_SECTION_POST_(?P<section>.*)",
    "VS_GLOBAL_SECTION_PRE_(?P<section>.*)",
    "VS_INCLUDE_IN_VSIX",
    "VS_IOT_EXTENSIONS_VERSION",
    "VS_IOT_STARTUP_TASK",
    "VS_KEYWORD",
    "VS_MOBILE_EXTENSIONS_VERSION",
    "VS_RESOURCE_GENERATOR",
    "VS_SCC_AUXPATH",
    "VS_SCC_LOCALPATH",
    "VS_SCC_PROJECTNAME",
    "VS_SCC_PROVIDER",
    "VS_SDK_REFERENCES",
    "VS_SHADER_ENTRYPOINT",
    "VS_SHADER_FLAGS",
    "VS_SHADER_MODEL",
    "VS_SHADER_OUTPUT_HEADER_FILE",
    "VS_SHADER_TYPE",
    "VS_SHADER_VARIABLE_NAME",
    "VS_STARTUP_PROJECT",
    "VS_TOOL_OVERRIDE",
    "VS_USER_PROPS",
    "VS_WINDOWS_TARGET_PLATFORM_MIN_VERSION",
    "VS_WINRT_COMPONENT",
    "VS_WINRT_EXTENSIONS",
    "VS_WINRT_REFERENCES",
    "VS_XAML_TYPE",
    "WILL_FAIL",
    "WIN32_EXECUTABLE",
    "WINDOWS_EXPORT_ALL_SYMBOLS",
    "WORKING_DIRECTORY",
    "WRAP_EXCLUDE",
    "XCODE_ATTRIBUTE_(?P<an_attribute>.*)",
    "XCODE_EMIT_EFFECTIVE_PLATFORM_NAME",
    "XCODE_EXPLICIT_FILE_TYPE",
    "XCODE_EXPLICIT_FILE_TYPE",
    "XCODE_FILE_ATTRIBUTES",
    "XCODE_LAST_KNOWN_FILE_TYPE",
    "XCODE_PRODUCT_TYPE",
    "XCTEST",
]


def stripped_patterns():
  regex = re.compile(r"\?P<[\w_]+>")
  return [regex.sub("", pattern) for pattern in PATTERNS]


CASE_SENSITIVE_REGEX = re.compile(
    "|".join(stripped_patterns()))
CASE_INSENSITIVE_REGEX = re.compile(
    "|".join(stripped_patterns()), re.IGNORECASE)



"""
Extract release notes for the given tag
"""

from __future__ import absolute_import, unicode_literals

import argparse
import io
import logging
import os
import re
import sys

from packaging import version as versiontool

logger = logging.getLogger(__name__)

GENERIC_MESSAGE = """
This release was automatically generated by the release pipeline.
""".strip()


def iterate_until_version(infile, version):
  """Read lines from `infile` until we have parsed a version heading in the
  form of::

    ---------
    v##.##.##
    ---------

  at which point the next line from `infile` will be the first line after the
  heading."""

  history = []
  ruler = re.compile("^-+$")

  for line in infile:
    line = line.rstrip()
    history.append(line)

    if len(history) < 3:
      continue

    try:
      line_version = versiontool.parse(history[-2].strip().rstrip("v"))
    except versiontool.InvalidVersion:
      continue

    if (ruler.match(history[-3]) and
        ruler.match(history[-1]) and
        (version is None or
         line_version.base_version == version.base_version)):
      break

    if len(history) > 3:
      for buffered_line in history[:-3]:
        yield buffered_line
      history = history[-3:]

  for buffered_line in history:
    yield buffered_line


def get_note_text(infile_path, versionstr):
  try:
    version = versiontool.parse(versionstr)
  except versiontool.InvalidVersion:
    return GENERIC_MESSAGE

  with io.open(infile_path, "r", encoding="utf-8") as infile:
    for _ in iterate_until_version(infile, version):
      pass

    lines = list(iterate_until_version(infile, None))
    content = "\n".join(lines[:-3])
    return re.sub(r"(\S)\n(\S)", r"\1\2", content).strip()


def main():
  parser = argparse.ArgumentParser(description=__doc__)
  parser.add_argument("infile_path")
  parser.add_argument("version")
  parser.add_argument("-o", "--outfile-path", default="-")

  args = parser.parse_args()

  if args.outfile_path == "-":
    args.outfile_path = os.dup(sys.stdout.fileno())
  with io.open(args.outfile_path, "w", encoding="utf-8") as outfile:
    outfile.write(get_note_text(args.infile_path, args.version))
    outfile.write("\n")


if __name__ == "__main__":
  main()



# -*- coding: utf-8 -*-
"""
Parse cmake listfiles and emit an html file containing semantic and
syntactic annotations.
"""

from __future__ import unicode_literals

import io
import logging
import os

from cmakelang import lex
from cmakelang.parse.common import TreeNode


def dump_html(node, outfile):
  """
  Write to `outfile` an html annoted version of the listfile which has been
  parsed into the parse tree rooted at `node`
  """

  if isinstance(node, TreeNode):
    outfile.write('<span class="cmf-{}">'.format(node.node_type.name))
    for child in node.children:
      dump_html(child, outfile)
    outfile.write('</span>')
  elif isinstance(node, lex.Token):
    outfile.write('<span class="cmf-{}">'.format(node.type.name))
    outfile.write(node.spelling)
    outfile.write('</span>')


def get_html(node, fullpage=False):
  """
  Return a string containing html markup of the annoted listfile which has
  been parsed into the parse tree rooted at `node`.
  """

  outfile = io.StringIO()
  dump_html(node, outfile)
  content = outfile.getvalue()
  if not fullpage:
    return content

  tpl_kwargs = {"content": content}

  try:
    import jinja2
  except ImportError:
    logging.error(
        "Cannot import jinja. Please install jinja in your python environment"
        " to use the fullpage html renderer")
    return None

  thisdir = os.path.realpath(os.path.dirname(__file__))
  tpldir = os.path.join(thisdir, "templates")
  env = jinja2.Environment(
      loader=jinja2.FileSystemLoader(tpldir),
      autoescape=jinja2.select_autoescape(["html"])
  )

  stylesheet_path = os.path.join(tpldir, "style.css")

  with io.open(stylesheet_path, "r", encoding="utf-8") as infile:
    tpl_kwargs["stylesheet"] = infile.read()

  template = env.get_template("layout.html.tpl")
  return template.render(**tpl_kwargs)



# -*- coding: utf-8 -*-
# PYTHON_ARGCOMPLETE_OK
"""
Check cmake listfile for lint
"""
from __future__ import unicode_literals

import argparse
import io
import logging
import os
import sys

import cmakelang
from cmakelang import common
from cmakelang.format import __main__
from cmakelang import configuration
from cmakelang import lex
from cmakelang import parse

from cmakelang.lint import basic_checker
from cmakelang.lint import lint_util

logger = logging.getLogger(__name__)


def process_file(config, local_ctx, infile_content):
  """
  Parse the input cmake file, re-format it, and print to the output file.
  """

  if config.format.line_ending == 'auto':
    detected = __main__.detect_line_endings(infile_content)
    config = config.clone()
    config.set_line_ending(detected)

  checker = basic_checker.LintChecker(config, local_ctx)
  checker.check_basics(infile_content)
  tokens = lex.tokenize(infile_content)
  checker.check_tokens(tokens)

  parse_db = parse.funs.get_parse_db()
  parse_db.update(parse.funs.get_funtree(config.parse.fn_spec))
  ctx = parse.ParseContext(parse_db, local_ctx, config)
  parse_tree = parse.parse(tokens, ctx)
  parse_tree.build_ancestry()
  checker.check_parse_tree(parse_tree)


def setup_argparse(argparser):
  argparser.add_argument('-v', '--version', action='version',
                         version=cmakelang.__version__)
  argparser.add_argument(
      '-l', '--log-level', default="info",
      choices=["error", "warning", "info", "debug"])

  mutex = argparser.add_mutually_exclusive_group()
  mutex.add_argument('--dump-config', choices=['yaml', 'json', 'python'],
                     default=None, const='python', nargs='?',
                     help='If specified, print the default configuration to '
                          'stdout and exit')
  mutex.add_argument('-o', '--outfile-path', default=None,
                     help='Write errors to this file. '
                          'Default is stdout.')

  argparser.add_argument(
      "--no-help", action="store_false", dest="with_help",
      help="When used with --dump-config, will omit helptext comments in the"
           " output"
  )
  argparser.add_argument(
      "--no-default", action="store_false", dest="with_defaults",
      help="When used with --dump-config, will omit any unmodified "
           "configuration value."
  )

  argparser.add_argument(
      "--suppress-decorations", action="store_true",
      help="Suppress the file title decoration and summary statistics")

  argparser.add_argument(
      '-c', '--config-files', nargs='+',
      help='path to configuration file(s)')
  argparser.add_argument('infilepaths', nargs='*')

  configuration.Configuration().add_to_argparser(argparser)


USAGE_STRING = """
cmake-lint [-h]
           [--dump-config {yaml,json,python} | -o OUTFILE_PATH]
           [-c CONFIG_FILE]
           infilepath [infilepath ...]
"""


def inner_main():
  """Parse arguments, open files, start work."""
  logging.basicConfig(level=logging.INFO, format="%(levelname)s %(message)s")

  argparser = argparse.ArgumentParser(
      description=__doc__,
      formatter_class=argparse.RawDescriptionHelpFormatter,
      usage=USAGE_STRING)

  setup_argparse(argparser)
  try:
    import argcomplete
    argcomplete.autocomplete(argparser)
  except ImportError:
    pass
  args = argparser.parse_args()
  logging.getLogger().setLevel(getattr(logging, args.log_level.upper()))

  if args.dump_config:
    config_dict = __main__.get_config(os.getcwd(), args.config_files)
    __main__.dump_config(args, config_dict, sys.stdout)
    sys.exit(0)

  if args.outfile_path is None:
    args.outfile_path = '-'

  if '-' in args.infilepaths:
    assert len(args.infilepaths) == 1, \
        "You cannot mix stdin as an input with other input files"

  if args.outfile_path == '-':
    outfile = io.open(os.dup(sys.stdout.fileno()),
                      mode='w', encoding="utf-8", newline='')
  else:
    outfile = io.open(args.outfile_path, 'w', encoding="utf-8", newline='')

  global_ctx = lint_util.GlobalContext(outfile)
  returncode = 0
  argdict = __main__.get_argdict(args)

  for infile_path in args.infilepaths:
    # NOTE(josh): have to load config once for every file, because we may pick
    # up a new config file location for each path
    if infile_path == '-':
      config_dict = __main__.get_config(os.getcwd(), args.config_files)
    else:
      config_dict = __main__.get_config(infile_path, args.config_files)
    config_dict.update(argdict)

    cfg = configuration.Configuration(**config_dict)
    if infile_path == '-':
      infile_path = os.dup(sys.stdin.fileno())

    try:
      infile = io.open(
          infile_path, mode='r', encoding=cfg.encode.input_encoding, newline='')
    except (IOError, OSError):
      logger.error("Failed to open %s for read", infile_path)
      returncode = 1
      continue

    try:
      with infile:
        intext = infile.read()
    except UnicodeDecodeError:
      logger.error(
          "Unable to read %s as %s", infile_path, cfg.encode.input_encoding)
      returncode = 1
      continue

    local_ctx = global_ctx.get_file_ctx(infile_path, cfg)
    process_file(cfg, local_ctx, intext)
    if not args.suppress_decorations:
      outfile.write("{}\n{}\n".format(infile_path, "=" * len(infile_path)))
    local_ctx.writeout(outfile)
    if not args.suppress_decorations:
      outfile.write("\n")
    if local_ctx.has_lint():
      returncode = 1

  if not args.suppress_decorations:
    global_ctx.write_summary(outfile)
  outfile.close()
  return returncode


def main():
  try:
    return inner_main()
  except SystemExit:
    return 0
  except common.UserError as ex:
    logger.fatal(ex.msg)
    return 1
  except common.InternalError as ex:
    logger.exception(ex.msg)
    return 2
  except AssertionError as ex:
    logger.exception(
        "An internal error occured. Please consider filing a bug report at "
        "github.com/cheshirekow/cmakelang/issues")
    return 2
  except:  # pylint: disable=bare-except
    logger.exception(
        "An internal error occured. Please consider filing a bug report at "
        "github.com/cheshirekow/cmakelang/issues")
    return 2


if __name__ == "__main__":
  sys.exit(main())



# -*- coding: utf-8 -*-
"""
Functions for parsing comments in markup
"""
from __future__ import unicode_literals
import math
import textwrap
import re
from cmakelang import common

# Matches comment strings like ``# TODO(josh):`` or ``# NOTE(josh):``
NOTE_REGEX = re.compile(r'^\s*[A-Z_]+\([^)]+\):.*')

# Matches comment lines that are clearly meant to separate sections or
# headers. The meaning of this regex is "a line consisting of three or more
# non-word characters ending with three or more non-word characters"
RULER_PATTERN = r'^\s*[^\w\s]{3}.*[^\w\s]{3}$'
RULER_REGEX = re.compile(RULER_PATTERN)

# Matches lines that start a bulleted list
BULLET_REGEX = re.compile(r'^(\s*)([\*-])( .+)$')

# Matches lines that start an itemized list
ENUM_REGEX = re.compile(r'^(\s*)\d+([.:])( .+)$')

# Matches a verbatim fence
FENCE_PATTERN = r'^\s*([`~]{3}[`~]*)(.*)$'
FENCE_REGEX = re.compile(FENCE_PATTERN)


class CommentType(common.EnumObject):
  _id_map = {}


CommentType.PARAGRAPH = CommentType(0)
CommentType.BULLET_LIST = CommentType(1)
CommentType.ENUM_LIST = CommentType(2)
CommentType.NOTE = CommentType(3)
CommentType.RULER = CommentType(4)
CommentType.SEPARATOR = CommentType(5)
CommentType.FENCE = CommentType(6)
CommentType.VERBATIM = CommentType(7)


class CommentItem(object):

  def __init__(self, kind):
    self.kind = kind
    self.indent = None
    self.lines = []

  def __repr__(self):
    return "{}".format(self.kind.name)


def parse(lines, config=None):
  """
  Parse comment lines. Returns objects of different formatable entities
  """
  # pylint: disable=too-many-statements

  obj_list = []
  state = None
  bullet_regex = None

  if config is None:
    fence_re = FENCE_REGEX
    ruler_re = RULER_REGEX
  else:
    fence_re = re.compile(config.markup.fence_pattern)
    ruler_re = re.compile(config.markup.ruler_pattern)

  for line in lines:
    fence_match = fence_re.match(line)
    if fence_match:
      obj_list.append(CommentItem(CommentType.FENCE))
      obj_list[-1].lines.append(fence_match.group(1).strip())
      content = fence_match.group(2).strip()
      line = content
      if state == CommentType.VERBATIM:
        state = None
      else:
        obj_list.append(CommentItem(CommentType.VERBATIM))
        state = CommentType.VERBATIM
      if not line:
        continue

    if state == CommentType.VERBATIM:
      if line and line[0] == ' ':
        obj_list[-1].lines.append(line[1:])
      else:
        obj_list[-1].lines.append(line)
      continue

    if not line:
      if state is CommentType.SEPARATOR:
        continue

      obj_list.append(CommentItem(CommentType.SEPARATOR))
      state = CommentType.SEPARATOR
      continue

    if ruler_re.match(line):
      obj_list.append(CommentItem(CommentType.RULER))
      obj_list[-1].lines.append(line.strip())
      state = CommentType.RULER
      continue

    if state in (None, CommentType.SEPARATOR, CommentType.RULER):
      match = BULLET_REGEX.match(line)
      if match:
        obj_list.append(CommentItem(CommentType.BULLET_LIST))
        indent_str = match.group(1)
        bullet_punctuation = match.group(2)
        obj_list[-1].lines.append(match.group(3))
        obj_list[-1].indent = len(indent_str)
        state = CommentType.BULLET_LIST

        if bullet_punctuation == '*':
          bullet_punctuation = r'\*'
        bullet_regex = re.compile(
            '^{}{}( .*)$'.format(indent_str, bullet_punctuation))
        continue

      match = ENUM_REGEX.match(line)
      if match:
        obj_list.append(CommentItem(CommentType.ENUM_LIST))
        indent_str = match.group(1)
        bullet_punctuation = match.group(2)
        obj_list[-1].lines.append(match.group(3).strip())
        obj_list[-1].indent = len(indent_str)
        state = CommentType.ENUM_LIST

        # TODO(josh) We want to match lines with either the same number of
        # spaces or with the colon in the same column
        bullet_regex = re.compile(
            r'^{}\d+{}( .*)$'.format(indent_str, bullet_punctuation))
        continue

      if NOTE_REGEX.match(line):
        obj_list.append(CommentItem(CommentType.NOTE))
        obj_list[-1].lines.append(line.strip())
        state = CommentType.NOTE
        continue

      if ruler_re.match(line):
        obj_list.append(CommentItem(CommentType.RULER))
        obj_list[-1].lines.append(line.strip())
        state = CommentType.RULER
        continue

      state = CommentType.PARAGRAPH
      obj_list.append(CommentItem(CommentType.PARAGRAPH))
      obj_list[-1].lines.append(line.strip())

    elif state in (CommentType.PARAGRAPH, CommentType.NOTE):
      if NOTE_REGEX.match(line):
        obj_list.append(CommentItem(CommentType.NOTE))
        state = CommentType.NOTE
      elif RULER_REGEX.match(line):
        obj_list.append(CommentItem(CommentType.RULER))
        state = CommentType.RULER
      obj_list[-1].lines.append(line.strip())

    elif state in (CommentType.BULLET_LIST, CommentType.ENUM_LIST):
      match = bullet_regex.match(line)
      if match:
        obj_list[-1].lines.append(match.group(1).strip())
      else:
        obj_list[-1].lines[-1] += '\n' + line.strip()

  return obj_list


def is_hashruler(item):
  """
  Return true if the markup item is a hash ruler, i.e.::

      ###########################
      # Like this ^^^ or this vvv
      ###########################
  """
  if item.kind != CommentType.RULER:
    return False
  if len(item.lines) != 1:
    return False
  if item.lines[0].strip('#'):
    return False
  return True


COMMON_KWARGS = dict(
    expand_tabs=True,
    replace_whitespace=True,
    drop_whitespace=True,
    break_long_words=False,
    break_on_hyphens=False
)


def format_item(config, line_width, item):
  """
  Return lines of formatted text based on the typeof markup
  """

  if item.kind == CommentType.SEPARATOR:
    return ['']
  if item.kind == CommentType.FENCE:
    return ['~~~']
  if item.kind == CommentType.VERBATIM:
    return [line.rstrip() for line in item.lines]
  if is_hashruler(item) and config.markup.canonicalize_hashrulers:
    return ['#' * line_width]
  if item.kind in (CommentType.PARAGRAPH, CommentType.NOTE, CommentType.RULER):
    wrapper = textwrap.TextWrapper(width=line_width, **COMMON_KWARGS)
    return common.stable_wrap(wrapper, '\n'.join(item.lines).strip())

  if item.kind == CommentType.BULLET_LIST:
    assert line_width > 2
    outlines = []

    wrapper = textwrap.TextWrapper(width=line_width - 2, **COMMON_KWARGS)
    for line in item.lines:
      increment_lines = common.stable_wrap(wrapper, line.strip())
      outlines.append(config.markup.bullet_char + ' ' + increment_lines[0])
      outlines.extend('  ' + iline for iline in increment_lines[1:])
    return outlines

  if item.kind == CommentType.ENUM_LIST:
    assert line_width > 2
    outlines = []
    wrapper = textwrap.TextWrapper(width=line_width - 2, **COMMON_KWARGS)

    digits = int(math.ceil(math.log(len(item.lines), 10)))
    fmt = '{:%dd}%s ' % (digits, config.markup.enum_char)
    indent = ' ' * (digits + 2)

    for idx, line in enumerate(item.lines):
      increment_lines = common.stable_wrap(wrapper, line.strip())
      outlines.append(fmt.format(idx + 1) + increment_lines[0])
      outlines.extend(indent + iline for iline in increment_lines[1:])
    return outlines

  raise AssertionError('Unexepected case')


def format_items(config, line_width, items):
  """
  Return lines of formatted text for the sequence of items within a comment
  block
  """

  outlines = []
  indent_history = []
  for item in items:
    if item.kind in (CommentType.BULLET_LIST, CommentType.ENUM_LIST):
      while indent_history and indent_history[-1] >= item.indent:
        indent_history.pop(-1)
      indent_history.append(item.indent)
      nindent = 2 * (len(indent_history) - 1)
      ilines = format_item(config, line_width - nindent, item)
      outlines.extend(' ' * nindent + iline for iline in ilines)
    else:
      outlines.extend(format_item(config, line_width, item))
      if item.kind != CommentType.SEPARATOR:
        indent_history = []
  return outlines



"""
Execute a GTK based program under broadway so that it can render without an
X backend.
"""

import argparse
import errno
import logging
import os
import subprocess
import sys

logger = logging.getLogger(__name__)


def pop_proc(procs, pid):
  for idx, proc in list(enumerate(procs)):
    if proc.pid == pid:
      return procs.pop(idx)
  raise ValueError("Unexpected pid: {}".format(pid))


def main():
  parser = argparse.ArgumentParser(description=__doc__)
  parser.add_argument(
      "-d", "--display-no", default=5, type=int,
      help="which display number the server should listen on")
  parser.add_argument(
      "--cwd", help="change to working directory to execute the program")
  parser.add_argument(
      "--executable",
      help="path to the binary to run, if the entry in argv[0] is different")
  parser.add_argument(
      "argv", nargs=argparse.REMAINDER)
  args = parser.parse_args()

  broadway = subprocess.Popen(["broadwayd", ":{:d}".format(args.display_no)])
  subenv = os.environ.copy()
  subenv["GDK_BACKEND"] = "broadway"
  subenv["BROADWAY_DISPLAY"] = ":{:d}".format(args.display_no)
  app = subprocess.Popen(
      args.argv, executable=args.executable, env=subenv, cwd=args.cwd)

  procs = [broadway, app]
  app_status = 1
  while procs:
    try:
      (pid, wstatus, _rusage) = os.wait4(-1, 0)
      pop_proc(procs, pid)
      if pid == app.pid:
        broadway.kill()
        app_status = wstatus

    except OSError as err:
      if err.errno == errno.EAGAIN:
        continue

      logger.exception("Error waiting for children")
      break

  if os.WIFEXITED(app_status):
    return os.WEXITSTATUS(app_status)
  if os.WIFSIGNALED(app_status):
    return -os.WSTOPSIG(app_status)
  return -1


if __name__ == "__main__":
  sys.exit(main())
